"""notification-daemon mock template

This creates the expected methods and properties of the notification-daemon
services, but no devices. You can specify non-default capabilities in
"parameters".
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = """
(c) 2012 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

BUS_NAME = "org.freedesktop.Notifications"
MAIN_OBJ = "/org/freedesktop/Notifications"
MAIN_IFACE = "org.freedesktop.Notifications"
SYSTEM_BUS = False

# default capabilities, can be modified with "capabilities" parameter
default_caps = [
    "body",
    "body-markup",
    "icon-static",
    "image/svg+xml",
    "private-synchronous",
    "append",
    "private-icon-only",
    "truncation",
]


def load(mock, parameters):
    caps = parameters["capabilities"].split() if "capabilities" in parameters else default_caps

    # next notification ID
    mock.next_id = 1

    mock.AddMethods(
        MAIN_IFACE,
        [
            ("GetCapabilities", "", "as", f"ret = {caps!r}"),
            (
                "CloseNotification",
                "i",
                "",
                "if args[0] < self.next_id: self.EmitSignal('', 'NotificationClosed', 'uu', [args[0], 1])",
            ),
            ("GetServerInformation", "", "ssss", 'ret = ("mock-notify", "test vendor", "1.0", "1.1")'),
            (
                "Notify",
                "susssasa{sv}i",
                "u",
                """if args[1]:
    ret = args[1]
else:
    ret = self.next_id
    self.next_id += 1
""",
            ),
        ],
    )



"""power-profiles-daemon < 0.20 mock template

This creates the expected methods and properties of the main
net.hadess.PowerProfiles object.

This provides only the non-deprecated D-Bus API as of version 0.9.
Note that this template is deprecated: Version 0.20 listens on a different
bus name/object path, it is provided in upower_power_profiles_daemon.py
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Bastien Nocera"
__copyright__ = """
(c) 2021, Red Hat Inc.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import dbus

BUS_NAME = "net.hadess.PowerProfiles"
MAIN_OBJ = "/net/hadess/PowerProfiles"
MAIN_IFACE = "net.hadess.PowerProfiles"
SYSTEM_BUS = True


def hold_profile(self, profile, reason, application_id):
    self.cookie += 1
    element = {"Profile": profile, "Reason": reason, "ApplicationId": application_id}
    self.holds[self.cookie] = element
    self.props[MAIN_IFACE]["ActiveProfileHolds"] = []
    for value in self.holds.values():
        self.props[MAIN_IFACE]["ActiveProfileHolds"].append(value)
    return self.cookie


def release_profile(self, cookie):
    self.holds.pop(cookie)
    self.props[MAIN_IFACE]["ActiveProfileHolds"] = []
    for value in self.holds.values():
        self.props[MAIN_IFACE]["ActiveProfileHolds"].append(value)
    if len(self.props[MAIN_IFACE]["ActiveProfileHolds"]) == 0:
        self.props[MAIN_IFACE]["ActiveProfileHolds"] = dbus.Array([], signature="(aa{sv})")


def load(mock, parameters):
    # Loaded!
    mock.loaded = True
    mock.cookie = 0
    mock.hold_profile = hold_profile
    mock.release_profile = release_profile
    mock.holds = {}

    props = {
        "ActiveProfile": parameters.get("ActiveProfile", "balanced"),
        "PerformanceDegraded": parameters.get("PerformanceDegraded", ""),
        "Profiles": [
            dbus.Dictionary({"Profile": "power-saver", "Driver": "dbusmock"}, signature="sv"),
            dbus.Dictionary({"Profile": "balanced", "Driver": "dbusmock"}, signature="sv"),
            dbus.Dictionary({"Profile": "performance", "Driver": "dbusmock"}, signature="sv"),
        ],
        "Actions": dbus.Array([], signature="s"),
        "ActiveProfileHolds": dbus.Array([], signature="(aa{sv})"),
    }
    mock.AddProperties(MAIN_IFACE, dbus.Dictionary(props, signature="sv"))

    mock.AddMethods(
        MAIN_IFACE,
        [
            ("HoldProfile", "sss", "u", "ret = self.hold_profile(self, args[0], args[1], args[2])"),
            ("ReleaseProfile", "u", "", "self.release_profile(self, args[0])"),
        ],
    )



"""sensors proxy mock template
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Marco Trevisan"
__copyright__ = """
(c) 2021 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import re

import dbus

from dbusmock import MOCK_IFACE

BUS_NAME = "net.hadess.SensorProxy"
MAIN_OBJ = "/net/hadess/SensorProxy"
MAIN_IFACE = "net.hadess.SensorProxy"
COMPASS_IFACE = "net.hadess.SensorProxy.Compass"
SYSTEM_BUS = True

CAMEL_TO_SNAKE_CASE_RE = re.compile(r"(?<!^)(?=[A-Z])")


def load(mock, parameters=None):
    mock.has_accelerometer = False
    mock.accelerometer_owners = {}
    mock.accelerometer_orientation = "undefined"
    mock.has_ambient_light = False
    mock.ambient_light_owners = {}
    mock.light_level_unit = "lux"
    mock.light_level = 0.0
    mock.has_proximity = False
    mock.proximity_near = False
    mock.proximity_owners = {}
    mock.has_compass = False
    mock.compass_owners = {}
    mock.compass_heading = -1.0

    if parameters:
        for p, v in parameters.items():
            setattr(mock, p, v)

    for iface in [MAIN_IFACE, COMPASS_IFACE]:
        mock.AddProperties(iface, mock.GetAll(iface))


def emit_signal_to_destination(mock, interface, name, signature, destination, *args):
    # We need to do this manually, could be made easier via
    # https://gitlab.freedesktop.org/dbus/dbus-python/-/merge_requests/13
    message = dbus.lowlevel.SignalMessage(mock.path, interface, name)
    if destination:
        message.set_destination(destination)
    message.append(*args, signature=signature)
    for location in mock.locations:
        location[0].send_message(message)


def emit_properties_changed(mock, interface=MAIN_IFACE, properties=None, destination=None):
    if properties is None:
        properties = mock.GetAll(interface)
    elif isinstance(properties, str):
        properties = [properties]

    if isinstance(properties, (list, set)):
        properties = {p: mock.Get(interface, p) for p in properties}
    elif not isinstance(properties, dict):
        raise TypeError("Unsupported properties type")

    emit_signal_to_destination(
        mock, dbus.PROPERTIES_IFACE, "PropertiesChanged", "sa{sv}as", destination, interface, properties, []
    )


@dbus.service.method(dbus.PROPERTIES_IFACE, in_signature="s", out_signature="a{sv}")
def GetAll(self, interface):
    if interface == MAIN_IFACE:
        return {
            "HasAccelerometer": dbus.Boolean(self.has_accelerometer),
            "AccelerometerOrientation": dbus.String(self.accelerometer_orientation),
            "HasAmbientLight": dbus.Boolean(self.has_ambient_light),
            "LightLevelUnit": dbus.String(self.light_level_unit),
            "LightLevel": dbus.Double(self.light_level),
            "HasProximity": dbus.Boolean(self.has_proximity),
            "ProximityNear": dbus.Boolean(self.proximity_near),
        }
    if interface == COMPASS_IFACE:
        return {
            "HasCompass": dbus.Boolean(self.has_compass),
            "CompassHeading": dbus.Double(self.compass_heading),
        }
    return dbus.Dictionary({}, signature="sv")


def register_owner(self, owners_dict, name):
    if name in owners_dict:
        return

    def name_cb(unique_name):
        if unique_name:
            return
        owners_dict.pop(name).cancel()

    owners_dict[name] = self.connection.watch_name_owner(name, name_cb)


def unregister_owner(owners_dict, name):
    watcher = owners_dict.pop(name, None)
    if watcher:
        watcher.cancel()


@dbus.service.method(MAIN_IFACE, sender_keyword="sender")
def ClaimAccelerometer(self, sender):
    register_owner(self, self.accelerometer_owners, sender)


@dbus.service.method(MAIN_IFACE, sender_keyword="sender")
def ReleaseAccelerometer(self, sender):
    unregister_owner(self.accelerometer_owners, sender)


@dbus.service.method(MAIN_IFACE, sender_keyword="sender")
def ClaimLight(self, sender):
    register_owner(self, self.ambient_light_owners, sender)


@dbus.service.method(MAIN_IFACE, sender_keyword="sender")
def ReleaseLight(self, sender):
    unregister_owner(self.ambient_light_owners, sender)


@dbus.service.method(MAIN_IFACE, sender_keyword="sender")
def ClaimProximity(self, sender):
    register_owner(self, self.proximity_owners, sender)


@dbus.service.method(MAIN_IFACE, sender_keyword="sender")
def ReleaseProximity(self, sender):
    unregister_owner(self.proximity_owners, sender)


@dbus.service.method(MAIN_IFACE, sender_keyword="sender")
def ClaimCompass(self, sender):
    register_owner(self, self.compass_owners, sender)


@dbus.service.method(MAIN_IFACE, sender_keyword="sender")
def ReleaseCompass(self, sender):
    unregister_owner(self.compass_owners, sender)


def sensor_to_attribute(sensor):
    if sensor == "light":
        return "ambient_light"
    return sensor


def is_valid_sensor_for_interface(sensor, interface):
    if interface == "net.hadess.SensorProxy":
        return sensor in ["accelerometer", "ambient_light", "proximity"]

    if interface == "net.hadess.SensorProxy.Compass":
        return sensor == "compass"

    return False


@dbus.service.method(MOCK_IFACE, in_signature="ssv")
def SetInternalProperty(self, interface, property_name, value):
    property_attribute = CAMEL_TO_SNAKE_CASE_RE.sub("_", property_name).lower()
    sensor = sensor_to_attribute(property_attribute.split("_")[0])

    owners = None
    if is_valid_sensor_for_interface(sensor, interface):
        if not getattr(self, f"has_{sensor}"):
            raise dbus.exceptions.DBusException(
                f"No {sensor} sensor available", name="org.freedesktop.DBus.Mock.Error"
            )

        owners = getattr(self, f"{sensor}_owners")
        # We allow setting a property from any client here, even if not claiming
        # but only owners, if any, will be notified about sensors changes

    pre_value = getattr(self, property_attribute)
    if pre_value != value:
        setattr(self, property_attribute, value)
        if owners:
            for owner in owners:
                emit_properties_changed(self, interface, property_name, owner)
        elif owners is None:
            emit_properties_changed(self, interface, property_name, None)


@dbus.service.method(MOCK_IFACE, in_signature="s")
def GetInternalProperty(self, property_name):
    property_attribute = CAMEL_TO_SNAKE_CASE_RE.sub("_", property_name).lower()
    value = getattr(self, property_attribute)

    if property_name.endswith("Owners"):
        return dbus.Array(value.keys(), signature="s")
    return value



"""systemd mock template
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Jonas Ådahl"
__copyright__ = """
(c) 2021 Red Hat
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import dbus
from gi.repository import GLib

from dbusmock import MOCK_IFACE, mockobject

BUS_PREFIX = "org.freedesktop.systemd1"
PATH_PREFIX = "/org/freedesktop/systemd1"


BUS_NAME = BUS_PREFIX
MAIN_OBJ = PATH_PREFIX
MAIN_IFACE = BUS_PREFIX + ".Manager"
UNIT_IFACE = BUS_PREFIX + ".Unit"
SYSTEM_BUS = True


def load(mock, _parameters):
    mock.next_job_id = 1
    mock.units = {}

    mock.AddProperties(MAIN_IFACE, {"Version": "v246"})


def escape_unit_name(name):
    for s in [".", "-"]:
        name = name.replace(s, "_")
    return name


def emit_job_new_remove(mock, job_id, job_path, name):
    mock.EmitSignal(MAIN_IFACE, "JobNew", "uos", [job_id, job_path, name])
    mock.EmitSignal(MAIN_IFACE, "JobRemoved", "uoss", [job_id, job_path, name, "done"])


@dbus.service.method(MAIN_IFACE, in_signature="ss", out_signature="o")
def StartUnit(self, name, _mode):
    job_id = self.next_job_id
    self.next_job_id += 1

    job_path = f"{PATH_PREFIX}/Job/{job_id}"
    GLib.idle_add(lambda: emit_job_new_remove(self, job_id, job_path, name))

    unit_path = self.units[str(name)]
    unit = mockobject.objects[unit_path]
    unit.UpdateProperties(UNIT_IFACE, {"ActiveState": "active"})

    return job_path


@dbus.service.method(MAIN_IFACE, in_signature="ssa(sv)a(sa(sv))", out_signature="o")
def StartTransientUnit(self, name, _mode, _properties, _aux):
    job_id = self.next_job_id
    self.next_job_id += 1

    job_path = f"{PATH_PREFIX}/Job/{job_id}"
    GLib.idle_add(lambda: emit_job_new_remove(self, job_id, job_path, name))

    return job_path


@dbus.service.method(MAIN_IFACE, in_signature="ss", out_signature="o")
def StopUnit(self, name, _mode):
    job_id = self.next_job_id
    self.next_job_id += 1

    job_path = f"{PATH_PREFIX}/Job/{job_id}"
    GLib.idle_add(lambda: emit_job_new_remove(self, job_id, job_path, name))

    unit_path = self.units[str(name)]
    unit = mockobject.objects[unit_path]
    unit.UpdateProperties(UNIT_IFACE, {"ActiveState": "inactive"})
    return job_path


@dbus.service.method(MAIN_IFACE, in_signature="s", out_signature="o")
def GetUnit(self, name):
    return self.units[str(name)]


@dbus.service.method(MOCK_IFACE, in_signature="s")
def AddMockUnit(self, name):
    unit_path = f"{PATH_PREFIX}/unit/{escape_unit_name(name)}"
    self.units[str(name)] = unit_path
    self.AddObject(
        unit_path,
        UNIT_IFACE,
        {
            "Id": name,
            "Names": [name],
            "LoadState": "loaded",
            "ActiveState": "inactive",
        },
        [],
    )



"""urfkill mock template

This creates the expected methods and properties of the main
urfkill object, but no devices. You can specify any property
such as urfkill in "parameters".
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Jussi Pakkanen"
__copyright__ = """
(C) 2015 Canonical ltd
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import dbus

import dbusmock

SYSTEM_BUS = True
BUS_NAME = "org.freedesktop.URfkill"
MAIN_OBJ = "/org/freedesktop/URfkill"

MAIN_IFACE = "org.freedesktop.URfkill"

individual_objects = ["BLUETOOTH", "FM", "GPS", "NFC", "UWB", "WIMAX", "WLAN", "WWAN"]
type2objectname = {
    1: "WLAN",
    2: "BLUETOOTH",
    3: "UWB",
    4: "WIMAX",
    5: "WWAN",
    6: "GPS",
    7: "FM",
}

KS_NOTAVAILABLE = -1
KS_UNBLOCKED = 0
KS_SOFTBLOCKED = 1
KS_HARDBLOCKED = 2


def toggle_flight_mode(self, new_block_state):
    new_block_state = bool(new_block_state)
    if self.flight_mode == new_block_state:
        return True
    self.flight_mode = new_block_state
    for i in individual_objects:
        old_value = self.internal_states[i]
        if old_value == 1:
            continue  # It was already blocked so we don't need to do anything
        path = "/org/freedesktop/URfkill/" + i
        obj = dbusmock.get_object(path)
        if new_block_state:
            obj.Set("org.freedesktop.URfkill.Killswitch", "state", 1)
            obj.EmitSignal("org.freedesktop.URfkill.Killswitch", "StateChanged", "", [])
        else:
            obj.Set("org.freedesktop.URfkill.Killswitch", "state", 0)
            obj.EmitSignal("org.freedesktop.URfkill.Killswitch", "StateChanged", "", [])
    self.EmitSignal(MAIN_IFACE, "FlightModeChanged", "b", [self.flight_mode])
    return True


def block(self, index, should_block):
    should_block = bool(should_block)
    if index not in type2objectname:
        return False
    objname = type2objectname[index]
    new_block_state = 1 if should_block else 0
    if self.internal_states[objname] != new_block_state:
        path = "/org/freedesktop/URfkill/" + objname
        obj = dbusmock.get_object(path)
        self.internal_states[objname] = new_block_state
        obj.Set("org.freedesktop.URfkill.Killswitch", "state", new_block_state)
        obj.EmitSignal("org.freedesktop.URfkill.Killswitch", "StateChanged", "", [])
    return True


def load(mock, parameters):
    mock.toggle_flight_mode = toggle_flight_mode
    mock.block = block
    mock.flight_mode = False
    mock.internal_states = {}
    for oname in individual_objects:
        mock.internal_states[oname] = KS_UNBLOCKED

    # First we create the main urfkill object.
    mock.AddMethods(
        MAIN_IFACE,
        [
            ("IsFlightMode", "", "b", "ret = self.flight_mode"),
            ("FlightMode", "b", "b", "ret = self.toggle_flight_mode(self, args[0])"),
            ("Block", "ub", "b", "ret = self.block(self, args[0], args[1])"),
        ],
    )

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "DaemonVersion": parameters.get("DaemonVersion", "0.6.0"),
                "KeyControl": parameters.get("KeyControl", True),
            },
            signature="sv",
        ),
    )

    for i in individual_objects:
        path = "/org/freedesktop/URfkill/" + i
        mock.AddObject(path, "org.freedesktop.URfkill.Killswitch", {"state": mock.internal_states[i]}, [])



# coding: UTF-8
"""Mock D-Bus objects for test suites."""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = """
(c) 2012 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""


from dbusmock.mockobject import MOCK_IFACE, OBJECT_MANAGER_IFACE, DBusMockObject, get_object, get_objects
from dbusmock.testcase import BusType, DBusTestCase, PrivateDBus, SpawnedMock

try:
    # created by setuptools_scm
    from dbusmock._version import __version__
except ImportError:
    __version__ = "0.git"


__all__ = [
    "MOCK_IFACE",
    "OBJECT_MANAGER_IFACE",
    "BusType",
    "DBusMockObject",
    "DBusTestCase",
    "PrivateDBus",
    "SpawnedMock",
    "get_object",
    "get_objects",
]



"""low-memory-monitor mock template

This creates the expected methods and properties of the main
org.freedesktop.LowMemoryMonitor object.

This provides only the 2.0 D-Bus API of low-memory-monitor.
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Bastien Nocera"
__copyright__ = """
(c) 2019, Red Hat Inc.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import dbus

from dbusmock import MOCK_IFACE

BUS_NAME = "org.freedesktop.LowMemoryMonitor"
MAIN_OBJ = "/org/freedesktop/LowMemoryMonitor"
MAIN_IFACE = "org.freedesktop.LowMemoryMonitor"
SYSTEM_BUS = True


def load(mock, _parameters):
    # Loaded!
    mock.loaded = True


@dbus.service.method(MOCK_IFACE, in_signature="y", out_signature="")
def EmitWarning(self, state):
    self.EmitSignal(MAIN_IFACE, "LowMemoryWarning", "y", [dbus.Byte(state)])



"""Mock templates for common D-Bus services"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.



"""bluetoothd mock template

This creates the expected methods and properties of the object manager
org.bluez object (/), the manager object (/org/bluez), but no adapters or
devices.

This supports BlueZ 5 only.
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Philip Withnall"
__copyright__ = """
(c) 2013 Collabora Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

from pathlib import Path

import dbus

from dbusmock import OBJECT_MANAGER_IFACE, mockobject

BUS_NAME = "org.bluez"
MAIN_OBJ = "/"
SYSTEM_BUS = True
IS_OBJECT_MANAGER = True

BLUEZ_MOCK_IFACE = "org.bluez.Mock"
AGENT_MANAGER_IFACE = "org.bluez.AgentManager1"
PROFILE_MANAGER_IFACE = "org.bluez.ProfileManager1"
ADAPTER_IFACE = "org.bluez.Adapter1"
MEDIA_IFACE = "org.bluez.Media1"
NETWORK_SERVER_IFACE = "org.bluez.Network1"
DEVICE_IFACE = "org.bluez.Device1"

# The device class of some arbitrary Android phone.
MOCK_PHONE_CLASS = 5898764


@dbus.service.method(AGENT_MANAGER_IFACE, in_signature="os", out_signature="")
def RegisterAgent(manager, agent_path, capability):
    all_caps = ["DisplayOnly", "DisplayYesNo", "KeyboardOnly", "NoInputNoOutput", "KeyboardDisplay"]

    if agent_path in manager.agent_paths:
        raise dbus.exceptions.DBusException(
            "Another agent is already registered " + manager.agent_path, name="org.bluez.Error.AlreadyExists"
        )

    if capability not in all_caps:
        raise dbus.exceptions.DBusException(
            "Unsupported capability " + capability, name="org.bluez.Error.InvalidArguments"
        )

    if not manager.default_agent:
        manager.default_agent = agent_path
    manager.agent_paths += [agent_path]
    manager.capabilities[str(agent_path)] = capability


@dbus.service.method(AGENT_MANAGER_IFACE, in_signature="o", out_signature="")
def UnregisterAgent(manager, agent_path):
    if agent_path not in manager.agent_paths:
        raise dbus.exceptions.DBusException("Agent not registered " + agent_path, name="org.bluez.Error.DoesNotExist")

    manager.agent_paths.remove(agent_path)
    del manager.capabilities[agent_path]
    if manager.default_agent == agent_path:
        if len(manager.agent_paths) > 0:
            manager.default_agent = manager.agent_paths[-1]
        else:
            manager.default_agent = None


@dbus.service.method(AGENT_MANAGER_IFACE, in_signature="o", out_signature="")
def RequestDefaultAgent(manager, agent_path):
    if agent_path not in manager.agent_paths:
        raise dbus.exceptions.DBusException("Agent not registered " + agent_path, name="org.bluez.Error.DoesNotExist")
    manager.default_agent = agent_path


def load(mock, _parameters):
    mock.AddObject(
        "/org/bluez",
        AGENT_MANAGER_IFACE,
        {},
        [
            ("RegisterAgent", "os", "", RegisterAgent),
            ("RequestDefaultAgent", "o", "", RequestDefaultAgent),
            ("UnregisterAgent", "o", "", UnregisterAgent),
        ],
    )

    bluez = mockobject.objects["/org/bluez"]
    bluez.AddMethods(
        PROFILE_MANAGER_IFACE,
        [
            ("RegisterProfile", "osa{sv}", "", ""),
            ("UnregisterProfile", "o", "", ""),
        ],
    )
    bluez.agent_paths = []
    bluez.capabilities = {}
    bluez.default_agent = None


@dbus.service.method(ADAPTER_IFACE, in_signature="o", out_signature="")
def RemoveDevice(adapter, path):
    adapter.RemoveObject(path)

    manager = mockobject.objects["/"]
    manager.EmitSignal(
        OBJECT_MANAGER_IFACE,
        "InterfacesRemoved",
        "oas",
        [
            dbus.ObjectPath(path),
            [DEVICE_IFACE],
        ],
    )


@dbus.service.method(ADAPTER_IFACE, in_signature="", out_signature="")
def StartDiscovery(adapter):
    adapter.props[ADAPTER_IFACE]["Discovering"] = True
    # NOTE: discovery filter support is minimal to mock
    # the Discoverable discovery filter
    if adapter.props[ADAPTER_IFACE]["DiscoveryFilter"] is not None:
        adapter.props[ADAPTER_IFACE]["Discoverable"] = True
    adapter.EmitSignal(
        dbus.PROPERTIES_IFACE,
        "PropertiesChanged",
        "sa{sv}as",
        [
            ADAPTER_IFACE,
            {
                "Discoverable": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discoverable"], variant_level=1),
                "Discovering": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discovering"], variant_level=1),
            },
            [],
        ],
    )


@dbus.service.method(ADAPTER_IFACE, in_signature="", out_signature="")
def StopDiscovery(adapter):
    adapter.props[ADAPTER_IFACE]["Discovering"] = False
    # NOTE: discovery filter support is minimal to mock
    # the Discoverable discovery filter
    if adapter.props[ADAPTER_IFACE]["DiscoveryFilter"] is not None:
        adapter.props[ADAPTER_IFACE]["Discoverable"] = False
    adapter.EmitSignal(
        dbus.PROPERTIES_IFACE,
        "PropertiesChanged",
        "sa{sv}as",
        [
            ADAPTER_IFACE,
            {
                "Discoverable": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discoverable"], variant_level=1),
                "Discovering": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discovering"], variant_level=1),
            },
            [],
        ],
    )


@dbus.service.method(ADAPTER_IFACE, in_signature="a{sv}", out_signature="")
def SetDiscoveryFilter(adapter, discovery_filter):
    adapter.props[ADAPTER_IFACE]["DiscoveryFilter"] = discovery_filter


@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="ss", out_signature="s")
def AddAdapter(self, device_name, system_name):
    """Convenience method to add a Bluetooth adapter

    You have to specify a device name which must be a valid part of an object
    path, e. g. "hci0", and an arbitrary system name (pretty hostname).

    Returns the new object path.
    """
    path = "/org/bluez/" + device_name
    address_start = int(device_name[-1])
    address = (
        f"{address_start:02d}:{address_start + 1:02d}:{address_start + 2:02d}:"
        f"{address_start + 3:02d}:{address_start + 4:02d}:{address_start + 5:02d}"
    )
    adapter_properties = {
        "UUIDs": dbus.Array(
            [
                # Reference:
                # http://git.kernel.org/cgit/bluetooth/bluez.git/tree/lib/uuid.h
                # PNP
                "00001200-0000-1000-8000-00805f9b34fb",
                # Generic Access Profile
                "00001800-0000-1000-8000-00805f9b34fb",
                # Generic Attribute Profile
                "00001801-0000-1000-8000-00805f9b34fb",
                # Audio/Video Remote Control Profile (remote)
                "0000110e-0000-1000-8000-00805f9b34fb",
                # Audio/Video Remote Control Profile (target)
                "0000110c-0000-1000-8000-00805f9b34fb",
            ],
            variant_level=1,
        ),
        "Discoverable": dbus.Boolean(False, variant_level=1),
        "Discovering": dbus.Boolean(False, variant_level=1),
        "Pairable": dbus.Boolean(True, variant_level=1),
        "Powered": dbus.Boolean(True, variant_level=1),
        "Address": dbus.String(address, variant_level=1),
        "AddressType": dbus.String("public", variant_level=1),
        "Alias": dbus.String(system_name, variant_level=1),
        "Modalias": dbus.String("usb:v1D6Bp0245d050A", variant_level=1),
        "Name": dbus.String(system_name, variant_level=1),
        # Reference:
        # http://bluetooth-pentest.narod.ru/software/
        # bluetooth_class_of_device-service_generator.html
        "Class": dbus.UInt32(268, variant_level=1),  # Computer, Laptop
        "DiscoverableTimeout": dbus.UInt32(180, variant_level=1),
        "PairableTimeout": dbus.UInt32(0, variant_level=1),
    }

    self.AddObject(
        path,
        ADAPTER_IFACE,
        # Properties
        adapter_properties,
        # Methods
        [
            ("RemoveDevice", "o", "", RemoveDevice),
            ("StartDiscovery", "", "", StartDiscovery),
            ("StopDiscovery", "", "", StopDiscovery),
            ("SetDiscoveryFilter", "a{sv}", "", SetDiscoveryFilter),
        ],
    )

    adapter = mockobject.objects[path]
    adapter.AddMethods(
        MEDIA_IFACE,
        [
            ("RegisterEndpoint", "oa{sv}", "", ""),
            ("UnregisterEndpoint", "o", "", ""),
        ],
    )
    adapter.AddMethods(
        NETWORK_SERVER_IFACE,
        [
            ("Register", "ss", "", ""),
            ("Unregister", "s", "", ""),
        ],
    )

    manager = mockobject.objects["/"]
    manager.EmitSignal(
        OBJECT_MANAGER_IFACE,
        "InterfacesAdded",
        "oa{sa{sv}}",
        [
            dbus.ObjectPath(path),
            {ADAPTER_IFACE: adapter_properties},
        ],
    )

    return path


@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="s")
def RemoveAdapter(self, device_name):
    """Convenience method to remove a Bluetooth adapter"""
    path = "/org/bluez/" + device_name
    # We could remove the devices related to the adapters here, but
    # when bluez crashes, the InterfacesRemoved aren't necessarily sent
    # devices first, so in effect, our laziness is testing an edge case
    # in the clients
    self.RemoveObject(path)

    manager = mockobject.objects["/"]
    manager.EmitSignal(
        OBJECT_MANAGER_IFACE,
        "InterfacesRemoved",
        "oas",
        [
            dbus.ObjectPath(path),
            [ADAPTER_IFACE],
        ],
    )


@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="s")
def RemoveAdapterWithDevices(self, device_name):
    """Convenience method to remove a Bluetooth adapter and all
    the devices associated to it
    """
    adapter_path = "/org/bluez/" + device_name
    adapter = mockobject.objects[adapter_path]
    manager = mockobject.objects["/"]

    to_remove = []
    for path in mockobject.objects:
        if path.startswith(adapter_path + "/"):
            to_remove.append(path)

    for path in to_remove:
        adapter.RemoveObject(path)
        manager.EmitSignal(
            OBJECT_MANAGER_IFACE,
            "InterfacesRemoved",
            "oas",
            [
                dbus.ObjectPath(path),
                [DEVICE_IFACE],
            ],
        )

    self.RemoveObject(adapter_path)
    manager.EmitSignal(
        OBJECT_MANAGER_IFACE,
        "InterfacesRemoved",
        "oas",
        [
            dbus.ObjectPath(adapter_path),
            [ADAPTER_IFACE],
        ],
    )


@dbus.service.method(DEVICE_IFACE, in_signature="", out_signature="")
def Pair(device):
    if device.paired:
        raise dbus.exceptions.DBusException("Device already paired", name="org.bluez.Error.AlreadyExists")
    device_address = device.props[DEVICE_IFACE]["Address"]
    adapter_device_name = Path(device.props[DEVICE_IFACE]["Adapter"]).name
    device.PairDevice(adapter_device_name, device_address)


@dbus.service.method(DEVICE_IFACE, in_signature="", out_signature="")
def Connect(device):
    if device.connected:
        raise dbus.exceptions.DBusException("Already Connected", name="org.bluez.Error.AlreadyConnected")
    device.connected = True
    device.EmitSignal(
        dbus.PROPERTIES_IFACE,
        "PropertiesChanged",
        "sa{sv}as",
        [
            DEVICE_IFACE,
            {
                "Connected": dbus.Boolean(device.connected, variant_level=1),
            },
            [],
        ],
    )


@dbus.service.method(DEVICE_IFACE, in_signature="", out_signature="")
def Disconnect(device):
    if not device.connected:
        raise dbus.exceptions.DBusException("Not Connected", name="org.bluez.Error.NotConnected")
    device.connected = False
    device.EmitSignal(
        dbus.PROPERTIES_IFACE,
        "PropertiesChanged",
        "sa{sv}as",
        [
            DEVICE_IFACE,
            {
                "Connected": dbus.Boolean(device.connected, variant_level=1),
            },
            [],
        ],
    )


@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="sss", out_signature="s")
def AddDevice(self, adapter_device_name, device_address, alias):
    """Convenience method to add a Bluetooth device

    You have to specify a device address which must be a valid Bluetooth
    address (e.g. 'AA:BB:CC:DD:EE:FF'). The alias is the human-readable name
    for the device (e.g. as set on the device itself), and the adapter device
    name is the device_name passed to AddAdapter.

    This will create a new, unpaired and unconnected device with some default properties
    like MOCK_PHONE_CLASS "Class" and a static "Modalias". Especially when working with more
    than one device, you may want to change these after creation.

    Returns the new object path.
    """
    device_name = "dev_" + device_address.replace(":", "_").upper()
    adapter_path = "/org/bluez/" + adapter_device_name
    path = adapter_path + "/" + device_name

    if adapter_path not in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Adapter {adapter_device_name} does not exist.", name=BLUEZ_MOCK_IFACE + ".NoSuchAdapter"
        )

    properties = {
        "Address": dbus.String(device_address, variant_level=1),
        "AddressType": dbus.String("public", variant_level=1),
        "Name": dbus.String(alias, variant_level=1),
        "Icon": dbus.String("phone", variant_level=1),
        "Class": dbus.UInt32(MOCK_PHONE_CLASS, variant_level=1),
        "Appearance": dbus.UInt16(0, variant_level=1),
        "UUIDs": dbus.Array([], signature="s", variant_level=1),
        "Paired": dbus.Boolean(False, variant_level=1),
        "Connected": dbus.Boolean(False, variant_level=1),
        "Trusted": dbus.Boolean(False, variant_level=1),
        "Blocked": dbus.Boolean(False, variant_level=1),
        "WakeAllowed": dbus.Boolean(False, variant_level=1),
        "Alias": dbus.String(alias, variant_level=1),
        "Adapter": dbus.ObjectPath(adapter_path, variant_level=1),
        "LegacyPairing": dbus.Boolean(False, variant_level=1),
        "Modalias": dbus.String("bluetooth:v000Fp1200d1436", variant_level=1),
        "RSSI": dbus.Int16(-79, variant_level=1),  # arbitrary
        "TxPower": dbus.Int16(0, variant_level=1),
        "ManufacturerData": dbus.Array([], signature="a{qv}", variant_level=1),
        "ServiceData": dbus.Array([], signature="a{sv}", variant_level=1),
        "ServicesResolved": dbus.Boolean(False, variant_level=1),
        "AdvertisingFlags": dbus.Array([], signature="ay", variant_level=1),
        "AdvertisingData": dbus.Array([], signature="a{yv}", variant_level=1),
    }

    self.AddObject(
        path,
        DEVICE_IFACE,
        # Properties
        properties,
        # Methods
        [
            ("CancelPairing", "", "", ""),
            ("Connect", "", "", Connect),
            ("ConnectProfile", "s", "", ""),
            ("Disconnect", "", "", Disconnect),
            ("DisconnectProfile", "s", "", ""),
            ("Pair", "", "", Pair),
        ],
    )
    device = mockobject.objects[path]
    device.paired = False
    device.connected = False

    manager = mockobject.objects["/"]
    manager.EmitSignal(
        OBJECT_MANAGER_IFACE,
        "InterfacesAdded",
        "oa{sa{sv}}",
        [
            dbus.ObjectPath(path),
            {DEVICE_IFACE: properties},
        ],
    )

    return path


@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="ss", out_signature="")
def PairDevice(_self, adapter_device_name, device_address):
    """Convenience method to mark an existing device as paired.

    You have to specify a device address which must be a valid Bluetooth
    address (e.g. 'AA:BB:CC:DD:EE:FF'). The adapter device name is the
    device_name passed to AddAdapter.

    This unblocks the device if it was blocked.

    If the specified adapter or device doesn't exist, a NoSuchAdapter or
    NoSuchDevice error will be returned on the bus.

    Returns nothing.
    """
    device_name = "dev_" + device_address.replace(":", "_").upper()
    adapter_path = "/org/bluez/" + adapter_device_name
    device_path = adapter_path + "/" + device_name

    if adapter_path not in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Adapter {adapter_device_name} does not exist.", name=BLUEZ_MOCK_IFACE + ".NoSuchAdapter"
        )
    if device_path not in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Device {device_name} does not exist.", name=BLUEZ_MOCK_IFACE + ".NoSuchDevice"
        )

    device = mockobject.objects[device_path]
    device.paired = True

    # Based off pairing with an Android phone.
    uuids = [
        "00001105-0000-1000-8000-00805f9b34fb",
        "0000110a-0000-1000-8000-00805f9b34fb",
        "0000110c-0000-1000-8000-00805f9b34fb",
        "00001112-0000-1000-8000-00805f9b34fb",
        "00001115-0000-1000-8000-00805f9b34fb",
        "00001116-0000-1000-8000-00805f9b34fb",
        "0000111f-0000-1000-8000-00805f9b34fb",
        "0000112f-0000-1000-8000-00805f9b34fb",
        "00001200-0000-1000-8000-00805f9b34fb",
    ]

    device.UpdateProperties(
        DEVICE_IFACE,
        {
            "UUIDs": dbus.Array(uuids, variant_level=1),
            "Paired": dbus.Boolean(True, variant_level=1),
            "LegacyPairing": dbus.Boolean(True, variant_level=1),
            "Blocked": dbus.Boolean(False, variant_level=1),
        },
    )


@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="ss", out_signature="")
def BlockDevice(_self, adapter_device_name, device_address):
    """Convenience method to mark an existing device as blocked.

    You have to specify a device address which must be a valid Bluetooth
    address (e.g. 'AA:BB:CC:DD:EE:FF'). The adapter device name is the
    device_name passed to AddAdapter.

    This disconnects the device if it was connected.

    If the specified adapter or device doesn't exist, a NoSuchAdapter or
    NoSuchDevice error will be returned on the bus.

    Returns nothing.
    """
    device_name = "dev_" + device_address.replace(":", "_").upper()
    adapter_path = "/org/bluez/" + adapter_device_name
    device_path = adapter_path + "/" + device_name

    if adapter_path not in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Adapter {adapter_device_name} does not exist.", name=BLUEZ_MOCK_IFACE + ".NoSuchAdapter"
        )
    if device_path not in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Device {device_name} does not exist.", name=BLUEZ_MOCK_IFACE + ".NoSuchDevice"
        )

    device = mockobject.objects[device_path]

    device.props[DEVICE_IFACE]["Blocked"] = dbus.Boolean(True, variant_level=1)
    device.props[DEVICE_IFACE]["Connected"] = dbus.Boolean(False, variant_level=1)

    device.EmitSignal(
        dbus.PROPERTIES_IFACE,
        "PropertiesChanged",
        "sa{sv}as",
        [
            DEVICE_IFACE,
            {
                "Blocked": dbus.Boolean(True, variant_level=1),
                "Connected": dbus.Boolean(False, variant_level=1),
            },
            [],
        ],
    )


@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="ss", out_signature="")
def ConnectDevice(_self, adapter_device_name, device_address):
    """Convenience method to mark an existing device as connected.

    You have to specify a device address which must be a valid Bluetooth
    address (e.g. 'AA:BB:CC:DD:EE:FF'). The adapter device name is the
    device_name passed to AddAdapter.

    This unblocks the device if it was blocked.

    If the specified adapter or device doesn't exist, a NoSuchAdapter or
    NoSuchDevice error will be returned on the bus.

    Returns nothing.
    """
    device_name = "dev_" + device_address.replace(":", "_").upper()
    adapter_path = "/org/bluez/" + adapter_device_name
    device_path = adapter_path + "/" + device_name

    if adapter_path not in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Adapter {adapter_device_name} does not exist.", name=BLUEZ_MOCK_IFACE + ".NoSuchAdapter"
        )
    if device_path not in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Device {device_name} does not exist.", name=BLUEZ_MOCK_IFACE + ".NoSuchDevice"
        )

    device = mockobject.objects[device_path]

    device.props[DEVICE_IFACE]["Blocked"] = dbus.Boolean(False, variant_level=1)
    device.props[DEVICE_IFACE]["Connected"] = dbus.Boolean(True, variant_level=1)

    device.EmitSignal(
        dbus.PROPERTIES_IFACE,
        "PropertiesChanged",
        "sa{sv}as",
        [
            DEVICE_IFACE,
            {
                "Blocked": dbus.Boolean(False, variant_level=1),
                "Connected": dbus.Boolean(True, variant_level=1),
            },
            [],
        ],
    )


@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="ss", out_signature="")
def DisconnectDevice(_self, adapter_device_name, device_address):
    """Convenience method to mark an existing device as disconnected.

    You have to specify a device address which must be a valid Bluetooth
    address (e.g. 'AA:BB:CC:DD:EE:FF'). The adapter device name is the
    device_name passed to AddAdapter.

    This does not change the device's blocked status.

    If the specified adapter or device doesn't exist, a NoSuchAdapter or
    NoSuchDevice error will be returned on the bus.

    Returns nothing.
    """
    device_name = "dev_" + device_address.replace(":", "_").upper()
    adapter_path = "/org/bluez/" + adapter_device_name
    device_path = adapter_path + "/" + device_name

    if adapter_path not in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Adapter {adapter_device_name} does not exist.", name=BLUEZ_MOCK_IFACE + ".NoSuchAdapter"
        )
    if device_path not in mockobject.objects:
        raise dbus.exceptions.DBusException(
            f"Device {device_name} does not exist.", name=BLUEZ_MOCK_IFACE + ".NoSuchDevice"
        )

    device = mockobject.objects[device_path]

    device.props[DEVICE_IFACE]["Connected"] = dbus.Boolean(False, variant_level=1)

    device.EmitSignal(
        dbus.PROPERTIES_IFACE,
        "PropertiesChanged",
        "sa{sv}as",
        [
            DEVICE_IFACE,
            {
                "Connected": dbus.Boolean(False, variant_level=1),
            },
            [],
        ],
    )



"""obexd mock template

This creates the expected methods and properties of the object manager
org.bluez.obex object (/), the manager object (/org/bluez/obex), but no agents
or clients.

This supports BlueZ 5 only.
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Philip Withnall"
__copyright__ = """
(c) 2013 Collabora Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import tempfile
from pathlib import Path

import dbus

from dbusmock import OBJECT_MANAGER_IFACE, mockobject

BUS_NAME = "org.bluez.obex"
MAIN_OBJ = "/"
SYSTEM_BUS = False
IS_OBJECT_MANAGER = True

OBEX_MOCK_IFACE = "org.bluez.obex.Mock"
AGENT_MANAGER_IFACE = "org.bluez.AgentManager1"
CLIENT_IFACE = "org.bluez.obex.Client1"
SESSION_IFACE = "org.bluez.obex.Session1"
PHONEBOOK_ACCESS_IFACE = "org.bluez.obex.PhonebookAccess1"
TRANSFER_IFACE = "org.bluez.obex.Transfer1"
TRANSFER_MOCK_IFACE = "org.bluez.obex.transfer1.Mock"


def load(mock, _parameters):
    mock.AddObject(
        "/org/bluez/obex",
        AGENT_MANAGER_IFACE,
        {},
        [
            ("RegisterAgent", "os", "", ""),
            ("UnregisterAgent", "o", "", ""),
        ],
    )

    obex = mockobject.objects["/org/bluez/obex"]
    obex.AddMethods(
        CLIENT_IFACE,
        [
            ("CreateSession", "sa{sv}", "o", CreateSession),
            ("RemoveSession", "o", "", RemoveSession),
        ],
    )


@dbus.service.method(CLIENT_IFACE, in_signature="sa{sv}", out_signature="o")
def CreateSession(self, destination, args):
    """OBEX method to create a new transfer session.

    The destination must be the address of the destination Bluetooth device.
    The given arguments must be a map from well-known keys to values,
    containing at least the `Target` key, whose value must be `PBAP` (other
    keys and values are accepted by the real daemon, but not by this mock
    daemon at the moment). If the target is missing or incorrect, an
    Unsupported error is returned on the bus.

    Returns the path of a new Session object.
    """

    if "Target" not in args or args["Target"].upper() != "PBAP":
        raise dbus.exceptions.DBusException(
            "Non-PBAP targets are not currently supported by this python-dbusmock template.",
            name=OBEX_MOCK_IFACE + ".Unsupported",
        )

    # Find the first unused session ID.
    client_path = "/org/bluez/obex/client"
    session_id = 0
    while client_path + "/session" + str(session_id) in mockobject.objects:
        session_id += 1

    path = client_path + "/session" + str(session_id)
    properties = {
        "Source": dbus.String("FIXME", variant_level=1),
        "Destination": dbus.String(destination, variant_level=1),
        "Channel": dbus.Byte(0, variant_level=1),
        "Target": dbus.String("FIXME", variant_level=1),
        "Root": dbus.String("FIXME", variant_level=1),
    }

    self.AddObject(
        path,
        SESSION_IFACE,
        # Properties
        properties,
        # Methods
        [
            ("GetCapabilities", "", "s", ""),  # Currently a no-op
        ],
    )

    session = mockobject.objects[path]
    session.AddMethods(
        PHONEBOOK_ACCESS_IFACE,
        [
            ("Select", "ss", "", ""),  # Currently a no-op
            # Currently a no-op
            ("List", "a{sv}", "a(ss)", 'ret = dbus.Array(signature="(ss)")'),
            # Currently a no-op
            ("ListFilterFields", "", "as", 'ret = dbus.Array(signature="(s)")'),
            ("PullAll", "sa{sv}", "sa{sv}", PullAll),
            ("GetSize", "", "q", "ret = 0"),
        ],
    )

    manager = mockobject.objects["/"]
    manager.EmitSignal(
        OBJECT_MANAGER_IFACE,
        "InterfacesAdded",
        "oa{sa{sv}}",
        [
            dbus.ObjectPath(path),
            {SESSION_IFACE: properties},
        ],
    )

    return path


@dbus.service.method(CLIENT_IFACE, in_signature="o", out_signature="")
def RemoveSession(self, session_path):
    """OBEX method to remove an existing transfer session.

    This takes the path of the transfer Session object and removes it.
    """

    manager = mockobject.objects["/"]

    # Remove all the session's transfers.
    transfer_id = 0
    while session_path + "/transfer" + str(transfer_id) in mockobject.objects:
        transfer_path = session_path + "/transfer" + str(transfer_id)
        transfer_id += 1

        self.RemoveObject(transfer_path)

        manager.EmitSignal(
            OBJECT_MANAGER_IFACE,
            "InterfacesRemoved",
            "oas",
            [
                dbus.ObjectPath(transfer_path),
                [TRANSFER_IFACE],
            ],
        )

    # Remove the session itself.
    self.RemoveObject(session_path)

    manager.EmitSignal(
        OBJECT_MANAGER_IFACE,
        "InterfacesRemoved",
        "oas",
        [
            dbus.ObjectPath(session_path),
            [SESSION_IFACE, PHONEBOOK_ACCESS_IFACE],
        ],
    )


@dbus.service.method(PHONEBOOK_ACCESS_IFACE, in_signature="sa{sv}", out_signature="sa{sv}")
def PullAll(self, target_file, filters):
    """OBEX method to start a pull transfer of a phone book.

    This doesn't complete the transfer; code to mock up activating and
    completing the transfer must be provided by the test driver, as it is
    too complex and test-specific to put here.

    The target_file is the absolute path for a file which will have zero or
    more vCards, separated by new-line characters, written to it if the method
    is successful (and the transfer is completed). This target_file is actually
    emitted in a TransferCreated signal, which is a special part of the mock
    interface designed to be handled by the test driver, which should then
    populate that file and call UpdateStatus on the Transfer object. The test
    driver is responsible for deleting the file once the test is complete.

    The filters parameter is a map of filters to be applied to the results
    device-side before transmitting them back to the adapter.

    Returns a tuple containing the path for a new Transfer D-Bus object
    representing the transfer, and a map of the initial properties of that
    Transfer object.
    """

    # Find the first unused session ID.
    session_path = self.path
    transfer_id = 0
    while session_path + "/transfer" + str(transfer_id) in mockobject.objects:
        transfer_id += 1

    transfer_path = session_path + "/transfer" + str(transfer_id)

    # Create a new temporary file to transfer to.
    with tempfile.NamedTemporaryFile(suffix=".vcf", prefix="tmp-bluez5-obex-PullAll_", delete=False) as temp_file:
        filename = Path(temp_file.name).resolve()

    props = {
        "Status": dbus.String("queued", variant_level=1),
        "Session": dbus.ObjectPath(session_path, variant_level=1),
        "Name": dbus.String(target_file, variant_level=1),
        "Filename": dbus.String(filename, variant_level=1),
        "Transferred": dbus.UInt64(0, variant_level=1),
    }

    self.AddObject(
        transfer_path,
        TRANSFER_IFACE,
        # Properties
        props,
        # Methods
        [
            ("Cancel", "", "", ""),  # Currently a no-op
        ],
    )

    transfer = mockobject.objects[transfer_path]
    transfer.AddMethods(
        TRANSFER_MOCK_IFACE,
        [
            ("UpdateStatus", "b", "", UpdateStatus),
        ],
    )

    manager = mockobject.objects["/"]
    manager.EmitSignal(
        OBJECT_MANAGER_IFACE,
        "InterfacesAdded",
        "oa{sa{sv}}",
        [
            dbus.ObjectPath(transfer_path),
            {TRANSFER_IFACE: props},
        ],
    )

    # Emit a behind-the-scenes signal that a new transfer has been created.
    manager.EmitSignal(OBEX_MOCK_IFACE, "TransferCreated", "sa{sv}s", [transfer_path, filters, str(filename)])

    return (transfer_path, props)


@dbus.service.signal(OBEX_MOCK_IFACE, signature="sa{sv}s")
def TransferCreated(_self, _path, _filters, _transfer_filename):
    """Mock signal emitted when a new Transfer object is created.

    This is not part of the BlueZ OBEX interface; it is purely for use by test
    driver code. It is emitted by the PullAll method, and is intended to be
    used as a signal to call UpdateStatus on the newly created Transfer
    (potentially after a timeout).

    The path is of the new Transfer object, and the filters are as provided to
    PullAll. The transfer filename is the full path and filename of a newly
    created empty temporary file which the test driver should write to.

    The test driver should then call UpdateStatus on the Transfer object each
    time a write is made to the transfer file. The final call to UpdateStatus
    should mark the transfer as complete.

    The test driver is responsible for deleting the transfer file once the test
    is complete.

    FIXME: Ideally the UpdateStatus method would only be used for completion,
    and all intermediate updates would be found by watching the size of the
    transfer file. However, that means adding a dependency on an inotify
    package, which seems a little much.
    """


@dbus.service.method(TRANSFER_MOCK_IFACE, in_signature="b", out_signature="")
def UpdateStatus(self, is_complete):
    """Mock method to update the transfer status.

    If is_complete is False, this marks the transfer is active; otherwise it
    marks the transfer as complete. It is an error to call this method after
    calling it with is_complete as True.

    In both cases, it updates the number of bytes transferred to be the current
    size of the transfer file (whose filename was emitted in the
    TransferCreated signal).
    """
    status = "complete" if is_complete else "active"
    transferred = Path(self.props[TRANSFER_IFACE]["Filename"]).stat().st_size

    self.props[TRANSFER_IFACE]["Status"] = status
    self.props[TRANSFER_IFACE]["Transferred"] = dbus.UInt64(transferred, variant_level=1)

    self.EmitSignal(
        dbus.PROPERTIES_IFACE,
        "PropertiesChanged",
        "sa{sv}as",
        [
            TRANSFER_IFACE,
            {
                "Status": dbus.String(status, variant_level=1),
                "Transferred": dbus.UInt64(transferred, variant_level=1),
            },
            [],
        ],
    )



"""NetworkManager mock template

This creates the expected methods and properties of the main
org.freedesktop.NetworkManager object, but no devices. You can specify any
property such as 'NetworkingEnabled', or 'WirelessEnabled' etc. in
"parameters".
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Iftikhar Ahmad"
__copyright__ = """
(c) 2012 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import binascii
import uuid

import dbus

import dbusmock
from dbusmock import MOCK_IFACE

BUS_NAME = "org.freedesktop.NetworkManager"
MAIN_OBJ = "/org/freedesktop"
MANAGER_IFACE = "org.freedesktop.NetworkManager"
MANAGER_OBJ = "/org/freedesktop/NetworkManager"
SETTINGS_OBJ = "/org/freedesktop/NetworkManager/Settings"
SETTINGS_IFACE = "org.freedesktop.NetworkManager.Settings"
DEVICE_IFACE = "org.freedesktop.NetworkManager.Device"
WIRELESS_DEVICE_IFACE = "org.freedesktop.NetworkManager.Device.Wireless"
ACCESS_POINT_IFACE = "org.freedesktop.NetworkManager.AccessPoint"
CSETTINGS_IFACE = "org.freedesktop.NetworkManager.Settings.Connection"
ACTIVE_CONNECTION_IFACE = "org.freedesktop.NetworkManager.Connection.Active"
SYSTEM_BUS = True
IS_OBJECT_MANAGER = True


# these really want to be dataclasses, but need to drop support for Python 3.6 for that
class NMState:  # pylint: disable=too-few-public-methods
    """Global state

    As per https://developer.gnome.org/NetworkManager/unstable/nm-dbus-types.html#NMState
    """

    NM_STATE_UNKNOWN = 0
    NM_STATE_ASLEEP = 10
    NM_STATE_DISCONNECTED = 20
    NM_STATE_DISCONNECTING = 30
    NM_STATE_CONNECTING = 40
    NM_STATE_CONNECTED_LOCAL = 50
    NM_STATE_CONNECTED_SITE = 60
    NM_STATE_CONNECTED_GLOBAL = 70


class NMConnectivityState:  # pylint: disable=too-few-public-methods
    """Connectvity state

    As per https://developer.gnome.org/NetworkManager/unstable/nm-dbus-types.html#NMConnectivityState
    """

    NM_CONNECTIVITY_UNKNOWN = 0
    NM_CONNECTIVITY_NONE = 1
    NM_CONNECTIVITY_PORTAL = 2
    NM_CONNECTIVITY_LIMITED = 3
    NM_CONNECTIVITY_FULL = 4


class NMActiveConnectionState:  # pylint: disable=too-few-public-methods
    """Active connection state

    As per https://developer.gnome.org/NetworkManager/unstable/nm-dbus-types.html#NMActiveConnectionState
    """

    NM_ACTIVE_CONNECTION_STATE_UNKNOWN = 0
    NM_ACTIVE_CONNECTION_STATE_ACTIVATING = 1
    NM_ACTIVE_CONNECTION_STATE_ACTIVATED = 2
    NM_ACTIVE_CONNECTION_STATE_DEACTIVATING = 3
    NM_ACTIVE_CONNECTION_STATE_DEACTIVATED = 4


class InfrastructureMode:  # pylint: disable=too-few-public-methods
    """Infrastructure mode

    As per https://developer.gnome.org/NetworkManager/unstable/nm-dbus-types.html#NM80211Mode
    """

    NM_802_11_MODE_UNKNOWN = 0
    NM_802_11_MODE_ADHOC = 1
    NM_802_11_MODE_INFRA = 2
    NM_802_11_MODE_AP = 3

    NAME_MAP = {
        NM_802_11_MODE_UNKNOWN: "unknown",
        NM_802_11_MODE_ADHOC: "adhoc",
        NM_802_11_MODE_INFRA: "infrastructure",
        NM_802_11_MODE_AP: "access-point",
    }


class DeviceState:  # pylint: disable=too-few-public-methods
    """Device states

    As per https://developer.gnome.org/NetworkManager/unstable/nm-dbus-types.html#NMDeviceState
    """

    UNKNOWN = 0
    UNMANAGED = 10
    UNAVAILABLE = 20
    DISCONNECTED = 30
    PREPARE = 40
    CONFIG = 50
    NEED_AUTH = 60
    IP_CONFIG = 70
    IP_CHECK = 80
    SECONDARIES = 90
    ACTIVATED = 100
    DEACTIVATING = 110
    FAILED = 120


class NM80211ApSecurityFlags:  # pylint: disable=too-few-public-methods
    """Security flags

    As per https://developer.gnome.org/NetworkManager/unstable/nm-dbus-types.html#NM80211ApSecurityFlags
    """

    NM_802_11_AP_SEC_NONE = 0x00000000
    NM_802_11_AP_SEC_PAIR_WEP40 = 0x00000001
    NM_802_11_AP_SEC_PAIR_WEP104 = 0x00000002
    NM_802_11_AP_SEC_PAIR_TKIP = 0x00000004
    NM_802_11_AP_SEC_PAIR_CCMP = 0x00000008
    NM_802_11_AP_SEC_GROUP_WEP40 = 0x00000010
    NM_802_11_AP_SEC_GROUP_WEP104 = 0x00000020
    NM_802_11_AP_SEC_GROUP_TKIP = 0x00000040
    NM_802_11_AP_SEC_GROUP_CCMP = 0x00000080
    NM_802_11_AP_SEC_KEY_MGMT_PSK = 0x00000100
    NM_802_11_AP_SEC_KEY_MGMT_802_1X = 0x00000200

    NAME_MAP = {
        NM_802_11_AP_SEC_KEY_MGMT_PSK: {"key-mgmt": "wpa-psk", "auth-alg": "open"},
    }


class NM80211ApFlags:  # pylint: disable=too-few-public-methods
    """Device flags

    As per https://developer.gnome.org/NetworkManager/unstable/nm-dbus-types.html#NM80211ApFlags
    """

    NM_802_11_AP_FLAGS_NONE = 0x00000000
    NM_802_11_AP_FLAGS_PRIVACY = 0x00000001


def activate_connection(self, conn, dev, ap):
    # find a new name
    count = 0
    active_connections = dbusmock.get_object(MANAGER_OBJ).Get(MANAGER_IFACE, "ActiveConnections")
    while True:
        path = dbus.ObjectPath("/org/freedesktop/NetworkManager/ActiveConnection/" + str(count))
        if path not in active_connections:
            break
        count += 1

    state = dbus.UInt32(NMActiveConnectionState.NM_ACTIVE_CONNECTION_STATE_ACTIVATED)
    devices = []
    if str(dev) != "/":
        devices.append(dev)
    return dbus.ObjectPath(AddActiveConnection(self, devices, conn, ap, str(count), state))


def deactivate_connection(self, active_conn_path):
    NM = dbusmock.get_object(MANAGER_OBJ)

    for dev_path in NM.GetDevices():
        RemoveActiveConnection(self, dev_path, active_conn_path)


def add_and_activate_connection(self, conn_conf, dev, ap):
    name = ap.rsplit("/", 1)[1]
    RemoveWifiConnection(self, dev, "/org/freedesktop/NetworkManager/Settings/" + name)

    try:
        raw_ssid = "".join([chr(byte) for byte in conn_conf["802-11-wireless"]["ssid"]])
    except KeyError:
        raw_ssid = dbusmock.get_object(ap).props["org.freedesktop.NetworkManager.AccessPoint"]["Ssid"].decode()
    wifi_conn = dbus.ObjectPath(AddWiFiConnection(self, dev, name, raw_ssid, ""))

    active_conn = activate_connection(self, wifi_conn, dev, ap)

    return (wifi_conn, active_conn)


def get_device_by_ip_iface(_self, iface):
    NM = dbusmock.get_object(MANAGER_OBJ)
    for dev_path in NM.GetDevices():
        dev_obj = dbusmock.get_object(dev_path)
        interface = dev_obj.Get(DEVICE_IFACE, "Interface")
        if interface == iface:
            return dev_path

    return None


def set_networking_enabled(self, networking_enabled):
    if networking_enabled:
        SetGlobalConnectionState(self, NMState.NM_STATE_CONNECTING)
        SetGlobalConnectionState(self, NMState.NM_STATE_CONNECTED_LOCAL)
        SetGlobalConnectionState(self, NMState.NM_STATE_CONNECTED_GLOBAL)
    else:
        SetGlobalConnectionState(self, NMState.NM_STATE_DISCONNECTING)
        SetGlobalConnectionState(self, NMState.NM_STATE_DISCONNECTED)
        SetGlobalConnectionState(self, NMState.NM_STATE_ASLEEP)

    SetNetworkingEnabled(self, networking_enabled)


def load(mock, parameters):
    manager_props = {
        "ActiveConnections": dbus.Array([], signature="o"),
        "Devices": dbus.Array([], signature="o"),
        "NetworkingEnabled": parameters.get("NetworkingEnabled", True),
        "Connectivity": parameters.get("Connectivity", dbus.UInt32(NMConnectivityState.NM_CONNECTIVITY_FULL)),
        "State": parameters.get("State", dbus.UInt32(NMState.NM_STATE_CONNECTED_GLOBAL)),
        "Startup": False,
        "Version": parameters.get("Version", "0.9.6.0"),
        "WimaxEnabled": parameters.get("WimaxEnabled", True),
        "WimaxHardwareEnabled": parameters.get("WimaxHardwareEnabled", True),
        "WirelessEnabled": parameters.get("WirelessEnabled", True),
        "WirelessHardwareEnabled": parameters.get("WirelessHardwareEnabled", True),
        "WwanEnabled": parameters.get("WwanEnabled", False),
        "WwanHardwareEnabled": parameters.get("WwanHardwareEnabled", True),
    }
    manager_methods = [
        ("GetDevices", "", "ao", 'ret = [k for k in objects.keys() if "/Devices" in k]'),
        ("GetPermissions", "", "a{ss}", "ret = {}"),
        ("state", "", "u", f"ret = self.Get('{MANAGER_IFACE}', 'State')"),
        ("CheckConnectivity", "", "u", f"ret = self.Get('{MANAGER_IFACE}', 'Connectivity')"),
        ("ActivateConnection", "ooo", "o", "ret = self.activate_connection(self, args[0], args[1], args[2])"),
        ("DeactivateConnection", "o", "", "self.deactivate_connection(self, args[0])"),
        (
            "AddAndActivateConnection",
            "a{sa{sv}}oo",
            "oo",
            "ret = self.add_and_activate_connection(self, args[0], args[1], args[2])",
        ),
        ("GetDeviceByIpIface", "s", "o", "ret = self.get_device_by_ip_iface(self, args[0])"),
        ("Reload", "u", "", ""),
        ("Enable", "b", "", "self.set_networking_enabled(self, args[0])"),
    ]

    mock.AddObject(MANAGER_OBJ, MANAGER_IFACE, manager_props, manager_methods)
    mock.object_manager_emit_added(MANAGER_OBJ)

    obj = dbusmock.get_object(MANAGER_OBJ)
    obj.activate_connection = activate_connection
    obj.deactivate_connection = deactivate_connection
    obj.add_and_activate_connection = add_and_activate_connection
    obj.get_device_by_ip_iface = get_device_by_ip_iface
    obj.set_networking_enabled = set_networking_enabled

    settings_props = {"Hostname": "hostname", "CanModify": True, "Connections": dbus.Array([], signature="o")}
    settings_methods = [
        ("ListConnections", "", "ao", f"ret = self.Get('{SETTINGS_IFACE}', 'Connections')"),
        ("GetConnectionByUuid", "s", "o", "ret = self.SettingsGetConnectionByUuid(args[0])"),
        ("AddConnection", "a{sa{sv}}", "o", "ret = self.SettingsAddConnection(args[0])"),
        ("AddConnectionUnsaved", "a{sa{sv}}", "o", "ret = self.SettingsAddConnection(args[0])"),
        ("SaveHostname", "s", "", ""),
    ]
    mock.AddObject(SETTINGS_OBJ, SETTINGS_IFACE, settings_props, settings_methods)
    mock.object_manager_emit_added(SETTINGS_OBJ)


@dbus.service.method(MOCK_IFACE, in_signature="sssv", out_signature="")
def SetProperty(_self, path, iface, name, value):
    obj = dbusmock.get_object(path)
    obj.Set(iface, name, value)
    obj.EmitSignal(iface, "PropertiesChanged", "a{sv}", [{name: value}])


@dbus.service.method(MOCK_IFACE, in_signature="u", out_signature="")
def SetGlobalConnectionState(self, state):
    self.SetProperty(MANAGER_OBJ, MANAGER_IFACE, "State", dbus.UInt32(state, variant_level=1))
    self.EmitSignal(MANAGER_IFACE, "StateChanged", "u", [state])


@dbus.service.method(MOCK_IFACE, in_signature="u", out_signature="")
def SetConnectivity(self, connectivity):
    self.SetProperty(MANAGER_OBJ, MANAGER_IFACE, "Connectivity", dbus.UInt32(connectivity, variant_level=1))


@dbus.service.method(MOCK_IFACE, in_signature="b", out_signature="")
def SetNetworkingEnabled(self, networking_enabled):
    self.SetProperty(
        MANAGER_OBJ, MANAGER_IFACE, "NetworkingEnabled", dbus.Boolean(networking_enabled, variant_level=1)
    )


@dbus.service.method(MOCK_IFACE, in_signature="ss", out_signature="")
def SetDeviceActive(_self, device_path, active_connection_path):
    dev_obj = dbusmock.get_object(device_path)
    dev_obj.Set(DEVICE_IFACE, "ActiveConnection", dbus.ObjectPath(active_connection_path))
    old_state = dev_obj.Get(DEVICE_IFACE, "State")
    dev_obj.Set(DEVICE_IFACE, "State", dbus.UInt32(DeviceState.ACTIVATED))
    dev_obj.Set(DEVICE_IFACE, "StateReason", (dbus.UInt32(DeviceState.ACTIVATED), dbus.UInt32(0)))

    dev_obj.EmitSignal(
        DEVICE_IFACE, "StateChanged", "uuu", [dbus.UInt32(DeviceState.ACTIVATED), old_state, dbus.UInt32(1)]
    )


@dbus.service.method(MOCK_IFACE, in_signature="s", out_signature="")
def SetDeviceDisconnected(_self, device_path):
    dev_obj = dbusmock.get_object(device_path)
    dev_obj.Set(DEVICE_IFACE, "ActiveConnection", dbus.ObjectPath("/"))
    old_state = dev_obj.Get(DEVICE_IFACE, "State")
    dev_obj.Set(DEVICE_IFACE, "State", dbus.UInt32(DeviceState.DISCONNECTED))
    dev_obj.Set(DEVICE_IFACE, "StateReason", (dbus.UInt32(DeviceState.DISCONNECTED), dbus.UInt32(0)))

    dev_obj.EmitSignal(
        DEVICE_IFACE, "StateChanged", "uuu", [dbus.UInt32(DeviceState.DISCONNECTED), old_state, dbus.UInt32(1)]
    )


@dbus.service.method(MOCK_IFACE, in_signature="ssi", out_signature="s")
def AddEthernetDevice(self, device_name, iface_name, state):
    """Add an ethernet device.

    You have to specify device_name, device interface name (e. g. eth0), and
    state. You can use the predefined DeviceState values (e. g.
    DeviceState.ACTIVATED) or supply a numeric value. For valid state values
    please visit
    https://developer.gnome.org/NetworkManager/unstable/nm-dbus-types.html#NMDeviceState

    Please note that this does not set any global properties.

    Returns the new object path.
    """
    path = "/org/freedesktop/NetworkManager/Devices/" + device_name
    wired_props = {
        "Carrier": False,
        "HwAddress": dbus.String("78:DD:08:D2:3D:43"),
        "PermHwAddress": dbus.String("78:DD:08:D2:3D:43"),
        "Speed": dbus.UInt32(0),
    }
    self.AddObject(path, "org.freedesktop.NetworkManager.Device.Wired", wired_props, [])

    props = {
        "DeviceType": dbus.UInt32(1),
        "State": dbus.UInt32(state),
        "StateReason": (dbus.UInt32(state), dbus.UInt32(0)),
        "Interface": iface_name,
        "ActiveConnection": dbus.ObjectPath("/"),
        "AvailableConnections": dbus.Array([], signature="o"),
        "AutoConnect": False,
        "Managed": True,
        "Driver": "dbusmock",
        "IpInterface": "",
    }

    obj = dbusmock.get_object(path)
    obj.AddProperties(DEVICE_IFACE, props)

    self.object_manager_emit_added(path)

    NM = dbusmock.get_object(MANAGER_OBJ)
    devices = NM.Get(MANAGER_IFACE, "Devices")
    devices.append(path)
    NM.Set(MANAGER_IFACE, "Devices", devices)
    NM.EmitSignal("org.freedesktop.NetworkManager", "DeviceAdded", "o", [path])

    return path


@dbus.service.method(MOCK_IFACE, in_signature="ssi", out_signature="s")
def AddWiFiDevice(self, device_name, iface_name, state):
    """Add a WiFi Device.

    You have to specify device_name, device interface name (e. g.  wlan0) and
    state. You can use the predefined DeviceState values (e. g.
    DeviceState.ACTIVATED) or supply a numeric value. For valid state values,
    please visit
    https://developer.gnome.org/NetworkManager/unstable/nm-dbus-types.html#NMDeviceState

    Please note that this does not set any global properties.

    Returns the new object path.
    """

    path = "/org/freedesktop/NetworkManager/Devices/" + device_name
    self.AddObject(
        path,
        WIRELESS_DEVICE_IFACE,
        {
            "HwAddress": dbus.String("11:22:33:44:55:66"),
            "PermHwAddress": dbus.String("11:22:33:44:55:66"),
            "Bitrate": dbus.UInt32(5400),
            "Mode": dbus.UInt32(2),
            "WirelessCapabilities": dbus.UInt32(255),
            "AccessPoints": dbus.Array([], signature="o"),
        },
        [
            ("GetAccessPoints", "", "ao", "ret = self.access_points"),
            ("GetAllAccessPoints", "", "ao", "ret = self.access_points"),
            ("RequestScan", "a{sv}", "", ""),
        ],
    )

    dev_obj = dbusmock.get_object(path)
    dev_obj.access_points = []
    dev_obj.AddProperties(
        DEVICE_IFACE,
        {
            "ActiveConnection": dbus.ObjectPath("/"),
            "AvailableConnections": dbus.Array([], signature="o"),
            "AutoConnect": False,
            "Managed": True,
            "Driver": "dbusmock",
            "DeviceType": dbus.UInt32(2),
            "State": dbus.UInt32(state),
            "StateReason": (dbus.UInt32(state), dbus.UInt32(0)),
            "Interface": iface_name,
            "IpInterface": iface_name,
        },
    )

    self.object_manager_emit_added(path)

    NM = dbusmock.get_object(MANAGER_OBJ)
    devices = NM.Get(MANAGER_IFACE, "Devices")
    devices.append(path)
    NM.Set(MANAGER_IFACE, "Devices", devices)
    NM.EmitSignal("org.freedesktop.NetworkManager", "DeviceAdded", "o", [path])

    return path


@dbus.service.method(MOCK_IFACE, in_signature="ssssuuuyu", out_signature="s")
def AddAccessPoint(self, dev_path, ap_name, ssid, hw_address, mode, frequency, rate, strength, security):
    """Add an access point to an existing WiFi device.

    You have to specify WiFi Device path, Access Point object name,
    ssid, hw_address, mode, frequency, rate, strength and security.
    For valid access point property values, please visit
    https://developer.gnome.org/NetworkManager/unstable/gdbus-org.freedesktop.NetworkManager.AccessPoint.html

    Please note that this does not set any global properties.

    Returns the new object path.
    """
    dev_obj = dbusmock.get_object(dev_path)
    ap_path = "/org/freedesktop/NetworkManager/AccessPoint/" + ap_name
    if ap_path in dev_obj.access_points:
        raise dbus.exceptions.DBusException(
            f"Access point {ap_name} on device {dev_path} already exists", name=MANAGER_IFACE + ".AlreadyExists"
        )

    flags = NM80211ApFlags.NM_802_11_AP_FLAGS_PRIVACY
    if security == NM80211ApSecurityFlags.NM_802_11_AP_SEC_NONE:
        flags = NM80211ApFlags.NM_802_11_AP_FLAGS_NONE

    self.AddObject(
        ap_path,
        ACCESS_POINT_IFACE,
        {
            "Ssid": dbus.ByteArray(ssid.encode("UTF-8")),
            "HwAddress": dbus.String(hw_address),
            "Flags": dbus.UInt32(flags),
            "LastSeen": dbus.Int32(1),
            "Frequency": dbus.UInt32(frequency),
            "MaxBitrate": dbus.UInt32(rate),
            "Mode": dbus.UInt32(mode),
            "RsnFlags": dbus.UInt32(security),
            "WpaFlags": dbus.UInt32(security),
            "Strength": dbus.Byte(strength),
        },
        [],
    )
    self.object_manager_emit_added(ap_path)

    dev_obj.access_points.append(ap_path)

    aps = dev_obj.Get(WIRELESS_DEVICE_IFACE, "AccessPoints")
    aps.append(ap_path)
    dev_obj.Set(WIRELESS_DEVICE_IFACE, "AccessPoints", aps)

    dev_obj.EmitSignal(WIRELESS_DEVICE_IFACE, "AccessPointAdded", "o", [ap_path])

    return ap_path


@dbus.service.method(MOCK_IFACE, in_signature="ssss", out_signature="s")
def AddWiFiConnection(self, dev_path, connection_name, ssid_name, _key_mgmt):
    """Add an available connection to an existing WiFi device and access point.

    You have to specify WiFi Device path, Connection object name,
    SSID and key management.

    The SSID must match one of the previously created access points.

    Please note that this only sets minimal global properties in ipv{4,6}, no others.

    Returns the new object path.
    """

    dev_obj = dbusmock.get_object(dev_path)
    connection_path = "/org/freedesktop/NetworkManager/Settings/" + connection_name
    connections = dev_obj.Get(DEVICE_IFACE, "AvailableConnections")

    settings_obj = dbusmock.get_object(SETTINGS_OBJ)
    main_connections = settings_obj.ListConnections()

    ssid = ssid_name.encode("UTF-8")

    # Find the access point by ssid
    access_point = None
    access_points = dev_obj.access_points
    access_point_path = None
    for ap_path in access_points:
        ap = dbusmock.get_object(ap_path)
        if ap.Get(ACCESS_POINT_IFACE, "Ssid") == ssid:
            access_point = ap
            access_point_path = ap_path
            break

    if not access_point:
        raise dbus.exceptions.DBusException(
            f"Access point with SSID [{ssid_name}] could not be found", name=MANAGER_IFACE + ".DoesNotExist"
        )

    hw_address = access_point.Get(ACCESS_POINT_IFACE, "HwAddress")
    mode = access_point.Get(ACCESS_POINT_IFACE, "Mode")
    security = access_point.Get(ACCESS_POINT_IFACE, "WpaFlags")

    if connection_path in connections or connection_path in main_connections:
        raise dbus.exceptions.DBusException(
            f"Connection {connection_name} on device {dev_path} already exists", name=MANAGER_IFACE + ".AlreadyExists"
        )

    # Parse mac address string into byte array
    mac_bytes = binascii.unhexlify(hw_address.replace(":", ""))

    settings = {
        "802-11-wireless": {
            "seen-bssids": [hw_address],
            "ssid": dbus.ByteArray(ssid),
            "mac-address": dbus.ByteArray(mac_bytes),
            "mode": InfrastructureMode.NAME_MAP[mode],
        },
        "connection": {
            "timestamp": dbus.UInt64(1374828522),
            "type": "802-11-wireless",
            "id": ssid_name,
            "uuid": str(uuid.uuid4()),
        },
        "ipv4": {
            "method": "auto",
        },
        "ipv6": {
            "method": "auto",
        },
    }

    if security != NM80211ApSecurityFlags.NM_802_11_AP_SEC_NONE:
        settings["802-11-wireless"]["security"] = "802-11-wireless-security"
        settings["802-11-wireless-security"] = NM80211ApSecurityFlags.NAME_MAP[security]

    self.AddObject(
        connection_path,
        CSETTINGS_IFACE,
        {"Unsaved": False},
        [
            ("Delete", "", "", "self.ConnectionDelete(self)"),
            ("GetSettings", "", "a{sa{sv}}", "ret = self.ConnectionGetSettings(self)"),
            ("GetSecrets", "s", "a{sa{sv}}", "ret = self.ConnectionGetSecrets(self, args[0])"),
            ("Update", "a{sa{sv}}", "", "self.ConnectionUpdate(self, args[0])"),
        ],
    )
    self.object_manager_emit_added(connection_path)

    connection_obj = dbusmock.get_object(connection_path)
    connection_obj.settings = settings
    connection_obj.connection_path = connection_path
    connection_obj.ConnectionDelete = ConnectionDelete
    connection_obj.ConnectionGetSettings = ConnectionGetSettings
    connection_obj.ConnectionGetSecrets = ConnectionGetSecrets
    connection_obj.ConnectionUpdate = ConnectionUpdate

    connections.append(dbus.ObjectPath(connection_path))
    dev_obj.Set(DEVICE_IFACE, "AvailableConnections", connections)

    main_connections.append(connection_path)
    settings_obj.Set(SETTINGS_IFACE, "Connections", main_connections)

    settings_obj.EmitSignal(SETTINGS_IFACE, "NewConnection", "o", [access_point_path])

    return connection_path


@dbus.service.method(MOCK_IFACE, in_signature="assssu", out_signature="s")
def AddActiveConnection(self, devices, connection_device, specific_object, name, state):
    """Add an active connection to an existing WiFi device.

    You have to a list of the involved WiFi devices, the connection path,
    the access point path, ActiveConnection object name and connection
    state.

    Please note that this does not set any global properties.

    Returns the new object path.
    """

    conn_obj = dbusmock.get_object(connection_device)
    settings = conn_obj.settings
    conn_uuid = settings["connection"]["uuid"]
    conn_type = settings["connection"]["type"]
    conn_id = settings["connection"]["id"]

    device_objects = [dbus.ObjectPath(dev) for dev in devices]

    active_connection_path = "/org/freedesktop/NetworkManager/ActiveConnection/" + name
    self.AddObject(
        active_connection_path,
        ACTIVE_CONNECTION_IFACE,
        {
            "Devices": dbus.Array(device_objects, signature="o"),
            "Default6": False,
            "Default": True,
            "Type": conn_type,
            "Vpn": (conn_type == "vpn"),
            "Connection": dbus.ObjectPath(connection_device),
            "Master": dbus.ObjectPath("/"),
            "SpecificObject": dbus.ObjectPath(specific_object),
            "Uuid": conn_uuid,
            "State": dbus.UInt32(state),
            "Id": conn_id,
        },
        [],
    )

    for dev_path in devices:
        self.SetDeviceActive(dev_path, active_connection_path)

    self.object_manager_emit_added(active_connection_path)

    NM = dbusmock.get_object(MANAGER_OBJ)
    active_connections = NM.Get(MANAGER_IFACE, "ActiveConnections")
    active_connections.append(dbus.ObjectPath(active_connection_path))
    NM.SetProperty(MANAGER_OBJ, MANAGER_IFACE, "ActiveConnections", active_connections)

    return active_connection_path


@dbus.service.method(MOCK_IFACE, in_signature="ss", out_signature="")
def RemoveAccessPoint(self, dev_path, ap_path):
    """Remove the specified access point.

    You have to specify the device to remove the access point from, and the
    path of the access point.

    Please note that this does not set any global properties.
    """

    dev_obj = dbusmock.get_object(dev_path)

    aps = dev_obj.Get(WIRELESS_DEVICE_IFACE, "AccessPoints")
    aps.remove(ap_path)
    dev_obj.Set(WIRELESS_DEVICE_IFACE, "AccessPoints", aps)

    dev_obj.access_points.remove(ap_path)

    dev_obj.EmitSignal(WIRELESS_DEVICE_IFACE, "AccessPointRemoved", "o", [ap_path])

    self.object_manager_emit_removed(ap_path)
    self.RemoveObject(ap_path)


@dbus.service.method(MOCK_IFACE, in_signature="ss", out_signature="")
def RemoveWifiConnection(self, dev_path, connection_path):
    """Remove the specified WiFi connection.

    You have to specify the device to remove the connection from, and the
    path of the Connection.

    Please note that this does not set any global properties.
    """

    dev_obj = dbusmock.get_object(dev_path)
    settings_obj = dbusmock.get_object(SETTINGS_OBJ)

    connections = dev_obj.Get(DEVICE_IFACE, "AvailableConnections")
    main_connections = settings_obj.ListConnections()

    if connection_path not in connections and connection_path not in main_connections:
        return

    connections.remove(dbus.ObjectPath(connection_path))
    dev_obj.Set(DEVICE_IFACE, "AvailableConnections", connections)

    main_connections.remove(connection_path)
    settings_obj.Set(SETTINGS_IFACE, "Connections", main_connections)

    settings_obj.EmitSignal(SETTINGS_IFACE, "ConnectionRemoved", "o", [connection_path])

    connection_obj = dbusmock.get_object(connection_path)
    connection_obj.EmitSignal(CSETTINGS_IFACE, "Removed", "", [])

    self.object_manager_emit_removed(connection_path)
    self.RemoveObject(connection_path)


@dbus.service.method(MOCK_IFACE, in_signature="ss", out_signature="")
def RemoveActiveConnection(self, dev_path, active_connection_path):
    """Remove the specified ActiveConnection.

    You have to specify the device to remove the connection from, and the
    path of the ActiveConnection.

    Please note that this does not set any global properties.
    """
    self.SetDeviceDisconnected(dev_path)

    NM = dbusmock.get_object(MANAGER_OBJ)
    active_connections = NM.Get(MANAGER_IFACE, "ActiveConnections")

    if active_connection_path not in active_connections:
        return

    active_connections.remove(dbus.ObjectPath(active_connection_path))
    NM.SetProperty(MANAGER_OBJ, MANAGER_IFACE, "ActiveConnections", active_connections)

    self.object_manager_emit_removed(active_connection_path)
    self.RemoveObject(active_connection_path)


@dbus.service.method(SETTINGS_IFACE, in_signature="a{sa{sv}}", out_signature="o")
def SettingsAddConnection(self, connection_settings):
    """Add a connection.

    connection_settings is a String String Variant Map Map. See
    https://developer.gnome.org/NetworkManager/0.9/spec.html #type-String_String_Variant_Map_Map

    If you omit uuid, this method adds one for you.
    """

    if "uuid" not in connection_settings["connection"]:
        connection_settings["connection"]["uuid"] = str(uuid.uuid4())

    NM = dbusmock.get_object(MANAGER_OBJ)
    settings_obj = dbusmock.get_object(SETTINGS_OBJ)
    main_connections = settings_obj.ListConnections()

    # Mimic how NM names connections
    count = 0
    while True:
        connection_obj_path = dbus.ObjectPath(SETTINGS_OBJ + "/" + str(count))
        if connection_obj_path not in main_connections:
            break
        count += 1
    connection_path = str(connection_obj_path)

    self.AddObject(
        connection_path,
        CSETTINGS_IFACE,
        {"Unsaved": False},
        [
            ("Delete", "", "", "self.ConnectionDelete(self)"),
            ("GetSettings", "", "a{sa{sv}}", "ret = self.ConnectionGetSettings(self)"),
            ("GetSecrets", "s", "a{sa{sv}}", "ret = self.ConnectionGetSecrets(self, args[0])"),
            ("Update", "a{sa{sv}}", "", "self.ConnectionUpdate(self, args[0])"),
        ],
    )
    self.object_manager_emit_added(connection_path)

    connection_obj = dbusmock.get_object(connection_path)
    connection_obj.settings = connection_settings
    connection_obj.connection_path = connection_path
    connection_obj.ConnectionDelete = ConnectionDelete
    connection_obj.ConnectionGetSettings = ConnectionGetSettings
    connection_obj.ConnectionGetSecrets = ConnectionGetSecrets
    connection_obj.ConnectionUpdate = ConnectionUpdate

    main_connections.append(connection_path)
    settings_obj.Set(SETTINGS_IFACE, "Connections", main_connections)

    settings_obj.EmitSignal(SETTINGS_IFACE, "NewConnection", "o", [connection_path])

    auto_connect = False
    if "autoconnect" in connection_settings["connection"]:
        auto_connect = connection_settings["connection"]["autoconnect"]

    if auto_connect:
        dev = None
        devices = NM.GetDevices()

        # Grab the first device.
        if len(devices) > 0:
            dev = devices[0]

        if dev:
            activate_connection(NM, connection_path, dev, connection_path)

    return connection_path


@dbus.service.method(SETTINGS_IFACE, in_signature="s", out_signature="o")
def SettingsGetConnectionByUuid(self, conn_uuid):
    conns = self.ListConnections()
    for o in conns:
        self.conn = dbusmock.get_object(o)
        settings = self.conn.GetSettings()
        if settings["connection"]["uuid"] == conn_uuid:
            return o
    raise dbus.exceptions.DBusException("No connection with the UUID was found.")


def ConnectionUpdate(self, settings):
    """Update settings on a connection.

    settings is a String String Variant Map Map. See
    https://developer.gnome.org/NetworkManager/0.9/spec.html#type-String_String_Variant_Map_Map
    """
    connection_path = self.connection_path

    NM = dbusmock.get_object(MANAGER_OBJ)
    settings_obj = dbusmock.get_object(SETTINGS_OBJ)

    main_connections = settings_obj.ListConnections()

    if connection_path not in main_connections:
        raise dbus.exceptions.DBusException(
            f"Connection {connection_path} does not exist",
            name=MANAGER_IFACE + ".DoesNotExist",
        )

    # Take care not to overwrite the secrets
    for setting_name in settings:
        setting = settings[setting_name]
        for k in setting:
            if setting_name not in self.settings:
                self.settings[setting_name] = {}
            self.settings[setting_name][k] = setting[k]

    self.EmitSignal(CSETTINGS_IFACE, "Updated", "", [])

    auto_connect = False
    if "autoconnect" in settings["connection"]:
        auto_connect = settings["connection"]["autoconnect"]

    if auto_connect:
        dev = None
        devices = NM.GetDevices()

        # Grab the first device.
        if len(devices) > 0:
            dev = devices[0]

        if dev:
            activate_connection(NM, connection_path, dev, connection_path)

    return connection_path


def ConnectionGetSettings(self):
    # Deep copy the settings with the secrets stripped
    # out. (NOTE: copy.deepcopy doesn't work with dbus
    # types).
    settings = {}
    for setting_name in self.settings:
        setting = self.settings[setting_name]
        for k in setting:
            if k != "secrets":
                if setting_name not in settings:
                    settings[setting_name] = {}
                settings[setting_name][k] = setting[k]

    return settings


def ConnectionGetSecrets(self, setting):
    settings = self.settings[setting]

    if "secrets" in settings:
        secrets = {setting: {"secrets": settings["secrets"]}}
    else:
        secrets = {setting: {"secrets": {"no-secrets": True}}}

    return secrets


def ConnectionDelete(self):
    """Deletes a connection.

    This also
        * removes the deleted connection from any device,
        * removes any active connection(s) it might be associated with,
        * removes it from the Settings interface,
        * as well as deletes the object from the mock.

    Note: If this was the only active connection, we change the global
    connection state.
    """
    connection_path = self.connection_path

    NM = dbusmock.get_object(MANAGER_OBJ)
    settings_obj = dbusmock.get_object(SETTINGS_OBJ)

    # Find the associated active connection(s).
    active_connections = NM.Get(MANAGER_IFACE, "ActiveConnections")
    associated_active_connections = []
    for ac in active_connections:
        ac_obj = dbusmock.get_object(ac)
        ac_con = ac_obj.Get(ACTIVE_CONNECTION_IFACE, "Connection")
        if ac_con == connection_path:
            associated_active_connections.append(ac)

    # We found that the connection we are deleting are associated to all
    # active connections and subsequently set the global state to
    # disconnected.
    if len(active_connections) == len(associated_active_connections):
        self.SetGlobalConnectionState(NMState.NM_STATE_DISCONNECTED)

    # Remove the connection from all associated devices.
    # We also remove all associated active connections.
    for dev_path in NM.GetDevices():
        dev_obj = dbusmock.get_object(dev_path)
        connections = dev_obj.Get(DEVICE_IFACE, "AvailableConnections")

        for ac in associated_active_connections:
            NM.RemoveActiveConnection(dev_path, ac)

        if connection_path not in connections:
            continue

        connections.remove(dbus.ObjectPath(connection_path))
        dev_obj.Set(DEVICE_IFACE, "AvailableConnections", connections)

    # Remove the connection from the settings interface
    main_connections = settings_obj.ListConnections()
    if connection_path not in main_connections:
        return
    main_connections.remove(connection_path)
    settings_obj.Set(SETTINGS_IFACE, "Connections", main_connections)
    settings_obj.EmitSignal(SETTINGS_IFACE, "ConnectionRemoved", "o", [connection_path])

    # Remove the connection from the mock
    connection_obj = dbusmock.get_object(connection_path)
    connection_obj.EmitSignal(CSETTINGS_IFACE, "Removed", "", [])

    self.object_manager_emit_removed(connection_path)
    self.RemoveObject(connection_path)



"""pytest fixtures for DBusMock"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = "(c) 2023 Martin Pitt <martin@piware.de>"

from typing import Iterator

import pytest

from dbusmock.testcase import BusType, PrivateDBus


@pytest.fixture(scope="session")
def dbusmock_system() -> Iterator[PrivateDBus]:
    """Export the whole DBusTestCase as a fixture, with the system bus started"""

    with PrivateDBus(BusType.SYSTEM) as bus:
        yield bus


@pytest.fixture(scope="session")
def dbusmock_session() -> Iterator[PrivateDBus]:
    """Export the whole DBusTestCase as a fixture, with the session bus started"""

    with PrivateDBus(BusType.SESSION) as bus:
        yield bus



# coding: UTF-8
"""Main entry point for running mock server."""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = """
(c) 2012 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import argparse
import json
import os
import subprocess
import sys

import dbusmock.mockobject
import dbusmock.testcase


def parse_args():
    """Parse command line arguments"""

    parser = argparse.ArgumentParser(description="mock D-Bus object")
    parser.add_argument(
        "-s",
        "--system",
        action="store_true",
        help="put object(s) on system bus (default: session bus or template's SYSTEM_BUS flag)",
    )
    parser.add_argument(
        "--session",
        action="store_true",
        help="put object(s) on session bus (default without template; overrides template's SYSTEM_BUS flag)",
    )
    parser.add_argument(
        "-l",
        "--logfile",
        metavar="PATH",
        help="path of log file",
    )
    parser.add_argument(
        "-t",
        "--template",
        metavar="NAME",
        help="template to load (instead of specifying name, path, interface)",
    )
    parser.add_argument(
        "name",  # fmt: off
        metavar="NAME",
        nargs="?",
        help='D-Bus name to claim (e. g. "com.example.MyService") (if not using -t)',
    )
    parser.add_argument(
        "path",
        metavar="PATH",
        nargs="?",
        help="D-Bus object path for initial/main object (if not using -t)",
    )
    parser.add_argument(
        "interface",
        metavar="INTERFACE",
        nargs="?",
        help="main D-Bus interface name for initial object (if not using -t)",
    )
    parser.add_argument(
        "-m",
        "--is-object-manager",
        action="store_true",
        help="automatically implement the org.freedesktop.DBus.ObjectManager interface",
    )
    parser.add_argument(
        "-p",
        "--parameters",
        help="JSON dictionary of parameters to pass to the template",
    )
    parser.add_argument(
        "-e",
        "--exec",
        nargs=argparse.REMAINDER,
        help="Command to run in the mock environment",
    )

    arguments = parser.parse_args()

    if arguments.template:
        if arguments.name or arguments.path or arguments.interface:
            parser.error("--template and specifying NAME/PATH/INTERFACE are mutually exclusive")
    else:
        if not arguments.name or not arguments.path or not arguments.interface:
            parser.error("Not using a template, you must specify NAME, PATH, and INTERFACE")

    if arguments.system and arguments.session:
        parser.error("--system and --session are mutually exclusive")

    return arguments


if __name__ == "__main__":
    import ctypes

    import dbus.mainloop.glib
    import dbus.service

    args = parse_args()
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

    system_bus = args.system
    if args.template:
        module = dbusmock.mockobject.load_module(args.template)
        args.name = module.BUS_NAME
        args.path = module.MAIN_OBJ
        if not args.session and not args.system:
            system_bus = module.SYSTEM_BUS

        if hasattr(module, "IS_OBJECT_MANAGER"):
            args.is_object_manager = module.IS_OBJECT_MANAGER
        else:
            args.is_object_manager = False

        if args.is_object_manager and not hasattr(module, "MAIN_IFACE"):
            args.interface = dbusmock.mockobject.OBJECT_MANAGER_IFACE
        else:
            args.interface = module.MAIN_IFACE

    bus = dbusmock.testcase.DBusTestCase.get_dbus(system_bus)

    # quit mock when the bus is going down
    should_run = {True}
    bus.add_signal_receiver(
        should_run.pop,
        signal_name="Disconnected",
        path="/org/freedesktop/DBus/Local",
        dbus_interface="org.freedesktop.DBus.Local",
    )

    bus_name = dbus.service.BusName(args.name, bus, allow_replacement=True, replace_existing=True, do_not_queue=True)

    main_object = dbusmock.mockobject.DBusMockObject(
        bus_name, args.path, args.interface, {}, args.logfile, args.is_object_manager
    )

    parameters = None
    if args.parameters:
        try:
            parameters = json.loads(args.parameters)
        except ValueError as detail:
            sys.stderr.write(f"Malformed JSON given for parameters: {detail}\n")
            sys.exit(2)

        if not isinstance(parameters, dict):
            sys.stderr.write("JSON parameters must be a dictionary\n")
            sys.exit(2)

    if args.template:
        main_object.AddTemplate(args.template, parameters)

    libglib = ctypes.cdll.LoadLibrary("libglib-2.0.so.0")

    dbusmock.mockobject.objects[args.path] = main_object

    if args.exec:
        with subprocess.Popen(args.exec) as exec_proc:
            exit_status = set()

            @ctypes.CFUNCTYPE(None, ctypes.c_int, ctypes.c_int)
            def on_process_watch(_pid, status):
                """Check if the launched process is still alive"""
                if os.WIFEXITED(status):
                    exit_status.add(os.WEXITSTATUS(status))
                else:
                    exit_status.add(1)
                should_run.pop()

            libglib.g_child_watch_add(exec_proc.pid, on_process_watch)

            while should_run:
                libglib.g_main_context_iteration(None, True)

            try:
                exec_proc.terminate()
                exec_proc.wait()
            except ProcessLookupError:
                pass
            sys.exit(exit_status.pop() if exit_status else exec_proc.returncode)
    else:
        while should_run:
            libglib.g_main_context_iteration(None, True)



"""gnome-shell screensaver mock template

This creates the expected methods and properties of the
org.gnome.ScreenSaver object.
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Bastien Nocera"
__copyright__ = """
(c) 2013 Red Hat Inc.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

BUS_NAME = "org.gnome.ScreenSaver"
MAIN_OBJ = "/org/gnome/ScreenSaver"
MAIN_IFACE = "org.gnome.ScreenSaver"
SYSTEM_BUS = False


def load(mock, _parameters):
    mock.AddMethods(
        MAIN_IFACE,
        [
            ("GetActive", "", "b", "ret = self.is_active"),
            ("GetActiveTime", "", "u", "ret = 1"),
            # fmt: off
            ("SetActive", "b", "", 'self.is_active = args[0]; self.EmitSignal("", "ActiveChanged", "b", [self.is_active])'),
            ("Lock", "", "", "time.sleep(1); self.SetActive(True)"),
            ("ShowMessage", "sss", "", ""),
            ("SimulateUserActivity", "", "", ""),
        ],
    )

    # default state
    mock.is_active = False



"""upowerd mock template

This creates the expected methods and properties of the main
org.freedesktop.UPower object, but no devices. You can specify any property
such as 'OnLowBattery' or the return value of 'SuspendAllowed',
'HibernateAllowed', and 'GetCriticalAction' in "parameters".

This provides the 1.0 D-Bus API of upower.
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = """
(c) 2012, 2013 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import dbus

import dbusmock
from dbusmock import MOCK_IFACE, mockobject

BUS_NAME = "org.freedesktop.UPower"
MAIN_OBJ = "/org/freedesktop/UPower"
MAIN_IFACE = "org.freedesktop.UPower"
SYSTEM_BUS = True
DEVICE_IFACE = "org.freedesktop.UPower.Device"


def load(mock, parameters):
    mock.AddMethods(
        MAIN_IFACE,
        [
            (
                "EnumerateDevices",
                "",
                "ao",
                'ret = [k for k in objects.keys() if "/devices" in k and not k.endswith("/DisplayDevice")]',
            ),
        ],
    )

    props = dbus.Dictionary(
        {
            "DaemonVersion": parameters.get("DaemonVersion", "0.99"),
            "OnBattery": parameters.get("OnBattery", False),
            "LidIsPresent": parameters.get("LidIsPresent", True),
            "LidIsClosed": parameters.get("LidIsClosed", False),
            "LidForceSleep": parameters.get("LidForceSleep", True),
            "IsDocked": parameters.get("IsDocked", False),
        },
        signature="sv",
    )

    mock.AddMethods(
        MAIN_IFACE,
        [
            ("GetCriticalAction", "", "s", f'ret = "{parameters.get("GetCriticalAction", "HybridSleep")}"'),
            ("GetDisplayDevice", "", "o", 'ret = "/org/freedesktop/UPower/devices/DisplayDevice"'),
        ],
    )

    mock.p_display_dev = "/org/freedesktop/UPower/devices/DisplayDevice"

    # add Display device; for defined properties, see
    # http://cgit.freedesktop.org/upower/tree/src/org.freedesktop.UPower.xml
    mock.AddObject(
        mock.p_display_dev,
        DEVICE_IFACE,
        {
            "Type": dbus.UInt32(0, variant_level=1),
            "State": dbus.UInt32(0, variant_level=1),
            "Percentage": dbus.Double(0.0, variant_level=1),
            "Energy": dbus.Double(0.0, variant_level=1),
            "EnergyFull": dbus.Double(0.0, variant_level=1),
            "EnergyRate": dbus.Double(0.0, variant_level=1),
            "TimeToEmpty": dbus.Int64(0, variant_level=1),
            "TimeToFull": dbus.Int64(0, variant_level=1),
            "IsPresent": dbus.Boolean(False, variant_level=1),
            "IconName": dbus.String("", variant_level=1),
            # LEVEL_NONE
            "WarningLevel": dbus.UInt32(1, variant_level=1),
        },
        [
            ("Refresh", "", "", ""),
        ],
    )

    mock.AddProperties(MAIN_IFACE, props)


@dbus.service.method(MOCK_IFACE, in_signature="ss", out_signature="s")
def AddAC(self, device_name, model_name):
    """Convenience method to add an AC object

    You have to specify a device name which must be a valid part of an object
    path, e. g. "mock_ac", and an arbitrary model name.

    Please note that this does not set any global properties such as
    "on-battery".

    Returns the new object path.
    """
    path = "/org/freedesktop/UPower/devices/" + device_name
    self.AddObject(
        path,
        DEVICE_IFACE,
        {
            "PowerSupply": dbus.Boolean(True, variant_level=1),
            "Model": dbus.String(model_name, variant_level=1),
            "Online": dbus.Boolean(True, variant_level=1),
        },
        [],
    )
    self.EmitSignal(MAIN_IFACE, "DeviceAdded", "o", [path])
    return path


@dbus.service.method(MOCK_IFACE, in_signature="ssdx", out_signature="s")
def AddDischargingBattery(self, device_name, model_name, percentage, seconds_to_empty):
    """Convenience method to add a discharging battery object

    You have to specify a device name which must be a valid part of an object
    path, e. g. "mock_ac", an arbitrary model name, the charge percentage, and
    the seconds until the battery is empty.

    Please note that this does not set any global properties such as
    "on-battery".

    Returns the new object path.
    """
    path = "/org/freedesktop/UPower/devices/" + device_name
    self.AddObject(
        path,
        DEVICE_IFACE,
        {
            "PowerSupply": dbus.Boolean(True, variant_level=1),
            "IsPresent": dbus.Boolean(True, variant_level=1),
            "Model": dbus.String(model_name, variant_level=1),
            "Percentage": dbus.Double(percentage, variant_level=1),
            "TimeToEmpty": dbus.Int64(seconds_to_empty, variant_level=1),
            "EnergyFull": dbus.Double(100.0, variant_level=1),
            "Energy": dbus.Double(percentage, variant_level=1),
            # UP_DEVICE_STATE_DISCHARGING
            "State": dbus.UInt32(2, variant_level=1),
            # UP_DEVICE_KIND_BATTERY
            "Type": dbus.UInt32(2, variant_level=1),
        },
        [],
    )
    self.EmitSignal(MAIN_IFACE, "DeviceAdded", "o", [path])
    return path


@dbus.service.method(MOCK_IFACE, in_signature="ssdx", out_signature="s")
def AddChargingBattery(self, device_name, model_name, percentage, seconds_to_full):
    """Convenience method to add a charging battery object

    You have to specify a device name which must be a valid part of an object
    path, e. g. "mock_ac", an arbitrary model name, the charge percentage, and
    the seconds until the battery is full.

    Please note that this does not set any global properties such as
    "on-battery".

    Returns the new object path.
    """
    path = "/org/freedesktop/UPower/devices/" + device_name
    self.AddObject(
        path,
        DEVICE_IFACE,
        {
            "PowerSupply": dbus.Boolean(True, variant_level=1),
            "IsPresent": dbus.Boolean(True, variant_level=1),
            "Model": dbus.String(model_name, variant_level=1),
            "Percentage": dbus.Double(percentage, variant_level=1),
            "TimeToFull": dbus.Int64(seconds_to_full, variant_level=1),
            "EnergyFull": dbus.Double(100.0, variant_level=1),
            "Energy": dbus.Double(percentage, variant_level=1),
            # UP_DEVICE_STATE_CHARGING
            "State": dbus.UInt32(1, variant_level=1),
            # UP_DEVICE_KIND_BATTERY
            "Type": dbus.UInt32(2, variant_level=1),
        },
        [],
    )
    self.EmitSignal(MAIN_IFACE, "DeviceAdded", "o", [path])
    return path


@dbus.service.method(MOCK_IFACE, in_signature="uuddddxxbsu", out_signature="")
def SetupDisplayDevice(
    self,
    _type,
    state,
    percentage,
    energy,
    energy_full,
    energy_rate,
    time_to_empty,
    time_to_full,
    is_present,
    icon_name,
    warning_level,
):
    """Convenience method to configure DisplayDevice properties

    This calls Set() for all properties that the DisplayDevice is defined to
    have, and is shorter if you have to completely set it up instead of
    changing just one or two properties.
    """
    display_props = mockobject.objects[self.p_display_dev]
    display_props.Set(DEVICE_IFACE, "Type", dbus.UInt32(_type))
    display_props.Set(DEVICE_IFACE, "State", dbus.UInt32(state))
    display_props.Set(DEVICE_IFACE, "Percentage", percentage)
    display_props.Set(DEVICE_IFACE, "Energy", energy)
    display_props.Set(DEVICE_IFACE, "EnergyFull", energy_full)
    display_props.Set(DEVICE_IFACE, "EnergyRate", energy_rate)
    display_props.Set(DEVICE_IFACE, "TimeToEmpty", dbus.Int64(time_to_empty))
    display_props.Set(DEVICE_IFACE, "TimeToFull", dbus.Int64(time_to_full))
    display_props.Set(DEVICE_IFACE, "IsPresent", is_present)
    display_props.Set(DEVICE_IFACE, "IconName", icon_name)
    display_props.Set(DEVICE_IFACE, "WarningLevel", dbus.UInt32(warning_level))


@dbus.service.method(MOCK_IFACE, in_signature="oa{sv}", out_signature="")
def SetDeviceProperties(_self, object_path, properties):
    """Convenience method to Set a device's properties.

    object_path: the device to update
    properties: dictionary of keys to dbus variants.

    Changing this property will trigger the device's PropertiesChanged signal.
    """
    device = dbusmock.get_object(object_path)

    # set the properties
    for key, value in properties.items():
        device.Set(DEVICE_IFACE, key, value)


@dbus.service.method(MOCK_IFACE, in_signature="o", out_signature="")
def RemoveDevice(self, device_path):
    self.RemoveObject(device_path)
    self.EmitSignal(MAIN_IFACE, "DeviceRemoved", "o", [device_path])



"""polkitd mock template

This creates the expected methods and properties of the main
org.freedesktop.PolicyKit1 object. By default, all actions are rejected.  You
can call AllowUnknown() and SetAllowed() on the mock D-Bus interface to control
which actions are allowed.
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = """
(c) 2013-2021 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import time

import dbus

from dbusmock import MOCK_IFACE

BUS_NAME = "org.freedesktop.PolicyKit1"
MAIN_OBJ = "/org/freedesktop/PolicyKit1/Authority"
MAIN_IFACE = "org.freedesktop.PolicyKit1.Authority"
SYSTEM_BUS = True


def load(mock, _parameters):
    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "BackendName": "local",
                "BackendVersion": "0.8.15",
                "BackendFeatures": dbus.UInt32(1, variant_level=1),
            },
            signature="sv",
        ),
    )

    # default state
    mock.allow_unknown = False
    mock.allowed = []
    mock.delay = 0
    mock.simulate_hang = False
    mock.hanging_actions = []
    mock.hanging_calls = []


@dbus.service.method(
    MAIN_IFACE, in_signature="(sa{sv})sa{ss}us", out_signature="(bba{ss})", async_callbacks=("ok_cb", "_err_cb")
)
def CheckAuthorization(self, _subject, action_id, _details, _flags, _cancellation_id, ok_cb, _err_cb):
    time.sleep(self.delay)
    allowed = action_id in self.allowed or self.allow_unknown
    ret = (allowed, False, {"test": "test"})

    if self.simulate_hang or action_id in self.hanging_actions:
        self.hanging_calls.append((ok_cb, ret))
    else:
        ok_cb(ret)


@dbus.service.method(MAIN_IFACE, in_signature="(sa{sv})ss")
def RegisterAuthenticationAgent(_self, _subject, _locale, _object_path):
    pass


@dbus.service.method(MOCK_IFACE, in_signature="b", out_signature="")
def AllowUnknown(self, default):
    """Control whether unknown actions are allowed

    This controls the return value of CheckAuthorization for actions which were
    not explicitly allowed by SetAllowed().
    """
    self.allow_unknown = default


@dbus.service.method(MOCK_IFACE, in_signature="as", out_signature="")
def SetAllowed(self, actions):
    """Set allowed actions"""

    self.allowed = actions


@dbus.service.method(MOCK_IFACE, in_signature="d", out_signature="")
def SetDelay(self, delay):
    """Makes the CheckAuthorization() method to delay"""
    self.delay = delay


@dbus.service.method(MOCK_IFACE, in_signature="b", out_signature="")
def SimulateHang(self, hang):
    """Makes the CheckAuthorization() method to hang"""
    self.simulate_hang = hang


@dbus.service.method(MOCK_IFACE, in_signature="as", out_signature="")
def SimulateHangActions(self, actions):
    """Makes the CheckAuthorization() method to hang on such actions"""
    self.hanging_actions = actions


@dbus.service.method(MOCK_IFACE, in_signature="", out_signature="")
def ReleaseHangingCalls(self):
    """Calls all the hanging callbacks"""
    for cb, ret in self.hanging_calls:
        cb(ret)
    self.hanging_calls = []


@dbus.service.method(MOCK_IFACE, in_signature="", out_signature="b")
def HaveHangingCalls(self):
    """Check if we've hangling calls"""
    return len(self.hanging_calls)



"""ofonod D-Bus mock template"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Martin Pitt"
__copyright__ = """
(c) 2013 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""


import dbus

import dbusmock

BUS_NAME = "org.ofono"
MAIN_OBJ = "/"
MAIN_IFACE = "org.ofono.Manager"
SYSTEM_BUS = True

NOT_IMPLEMENTED = 'raise dbus.exceptions.DBusException("not implemented", name="org.ofono.Error.NotImplemented")'


#  interface org.ofono.Manager {
#    methods:
#      GetModems(out a(oa{sv}) modems);
#    signals:
#      ModemAdded(o path,
#                 a{sv} properties);
#      ModemRemoved(o path);
#  };

_parameters = {}


def load(mock, parameters):
    global _parameters  # pylint: disable=global-statement
    mock.modems = []  # object paths
    mock.modem_serial_counter = 0
    mock.imsi_counter = 0
    mock.iccid_counter = 0
    _parameters = parameters
    mock.AddMethod(
        MAIN_IFACE,
        "GetModems",
        "",
        "a(oa{sv})",
        'ret = [(m, objects[m].GetAll("org.ofono.Modem")) for m in self.modems]',
    )

    if not parameters.get("no_modem", False):
        mock.AddModem(parameters.get("ModemName", "ril_0"), {})


#  interface org.ofono.Modem {
#    methods:
#      GetProperties(out a{sv} properties);
#      SetProperty(in  s property,
#                  in  v value);
#    signals:
#      PropertyChanged(s name,
#                      v value);
#  };


@dbus.service.method(dbusmock.MOCK_IFACE, in_signature="sa{sv}", out_signature="s")
def AddModem(self, name, _properties):
    """Convenience method to add a modem

    You have to specify a device name which must be a valid part of an object
    path, e. g. "mock_ac". For future extensions you can specify a "properties"
    array, but no extra properties are supported for now.

    Returns the new object path.
    """
    path = "/" + name
    self.AddObject(
        path,
        "org.ofono.Modem",
        {
            "Online": dbus.Boolean(True, variant_level=1),
            "Powered": dbus.Boolean(True, variant_level=1),
            "Lockdown": dbus.Boolean(False, variant_level=1),
            "Emergency": dbus.Boolean(False, variant_level=1),
            "Manufacturer": dbus.String("Fakesys", variant_level=1),
            "Model": dbus.String("Mock Modem", variant_level=1),
            "Revision": dbus.String("0815.42", variant_level=1),
            "Serial": dbus.String(new_modem_serial(self), variant_level=1),
            "Type": dbus.String("hardware", variant_level=1),
            "Interfaces": [
                "org.ofono.CallVolume",
                "org.ofono.VoiceCallManager",
                "org.ofono.NetworkRegistration",
                "org.ofono.SimManager",
                # 'org.ofono.MessageManager',
                "org.ofono.ConnectionManager",
                # 'org.ofono.NetworkTime'
            ],
            # 'Features': ['sms', 'net', 'gprs', 'sim']
            "Features": ["gprs", "net"],
        },
        [
            ("GetProperties", "", "a{sv}", 'ret = self.GetAll("org.ofono.Modem")'),
            (
                "SetProperty",
                "sv",
                "",
                'self.Set("org.ofono.Modem", args[0], args[1]); '
                'self.EmitSignal("org.ofono.Modem", "PropertyChanged",'
                ' "sv", [args[0], args[1]])',
            ),
        ],
    )
    obj = dbusmock.mockobject.objects[path]
    obj.name = name
    add_voice_call_api(obj)
    add_netreg_api(obj)
    add_simmanager_api(self, obj)
    add_connectionmanager_api(obj)
    self.modems.append(path)
    props = obj.GetAll("org.ofono.Modem", dbus_interface=dbus.PROPERTIES_IFACE)
    self.EmitSignal(MAIN_IFACE, "ModemAdded", "oa{sv}", [path, props])
    return path


# Generate a new modem serial number so each modem we add gets a unique one.
# Use a counter so that the result is predictable for tests.
def new_modem_serial(mock):
    serial = f"12345678-1234-1234-1234-{mock.modem_serial_counter:012d}"
    mock.modem_serial_counter += 1
    return serial


# Generate a new unique IMSI (start with USA/AT&T 310/150 to match the MCC/MNC SIM properties)
# Use a counter so that the result is predictable for tests.
def new_imsi(mock):
    imsi = f"310150{mock.imsi_counter:09d}"
    mock.imsi_counter += 1
    return imsi


# Generate a new unique ICCID
# Use a counter so that the result is predictable for tests.
def new_iccid(mock):
    iccid = f"893581234{mock.iccid_counter:012d}"
    mock.iccid_counter += 1
    return iccid


#  interface org.ofono.VoiceCallManager {
#    methods:
#      GetProperties(out a{sv} properties);
#      Dial(in  s number,
#           in  s hide_callerid,
#           out o path);
#      Transfer();
#      SwapCalls();
#      ReleaseAndAnswer();
#      ReleaseAndSwap();
#      HoldAndAnswer();
#      HangupAll();
#      PrivateChat(in  o call,
#                  out ao calls);
#      CreateMultiparty(out o calls);
#      HangupMultiparty();
#      SendTones(in  s SendTones);
#      GetCalls(out a(oa{sv}) calls_with_properties);
#    signals:
#      Forwarded(s type);
#      BarringActive(s type);
#      PropertyChanged(s name,
#                      v value);
#      CallAdded(o path,
#                a{sv} properties);
#      CallRemoved(o path);
#  };


def add_voice_call_api(mock):
    """Add org.ofono.VoiceCallManager API to a mock"""

    # also add an emergency number which is not a real one, in case one runs a
    # test case against a production ofono :-)
    mock.AddProperty("org.ofono.VoiceCallManager", "EmergencyNumbers", ["911", "13373"])

    mock.calls = []  # object paths

    mock.AddMethods(
        "org.ofono.VoiceCallManager",
        [
            ("GetProperties", "", "a{sv}", 'ret = self.GetAll("org.ofono.VoiceCallManager")'),
            ("Transfer", "", "", ""),
            ("SwapCalls", "", "", ""),
            ("ReleaseAndAnswer", "", "", ""),
            ("ReleaseAndSwap", "", "", ""),
            ("HoldAndAnswer", "", "", ""),
            ("SendTones", "s", "", ""),
            ("PrivateChat", "o", "ao", NOT_IMPLEMENTED),
            ("CreateMultiparty", "", "o", NOT_IMPLEMENTED),
            ("HangupMultiparty", "", "", NOT_IMPLEMENTED),
            (
                "GetCalls",
                "",
                "a(oa{sv})",
                'ret = [(c, objects[c].GetAll("org.ofono.VoiceCall")) for c in self.calls]',
            ),
        ],
    )


@dbus.service.method("org.ofono.VoiceCallManager", in_signature="ss", out_signature="s")
def Dial(self, number, _hide_callerid):
    # pylint: disable=protected-access # _object_path
    path = f"{self._object_path}/voicecall{(len(self.calls) + 1):02}"
    self.AddObject(
        path,
        "org.ofono.VoiceCall",
        {
            "State": dbus.String("dialing", variant_level=1),
            "LineIdentification": dbus.String(number, variant_level=1),
            "Name": dbus.String("", variant_level=1),
            "Multiparty": dbus.Boolean(False, variant_level=1),
            "RemoteHeld": dbus.Boolean(False, variant_level=1),
            "RemoteMultiparty": dbus.Boolean(False, variant_level=1),
            "Emergency": dbus.Boolean(False, variant_level=1),
        },
        [
            ("GetProperties", "", "a{sv}", 'ret = self.GetAll("org.ofono.VoiceCall")'),
            ("Deflect", "s", "", NOT_IMPLEMENTED),
            (
                "Hangup",
                "",
                "",
                "self.parent.calls.remove(self._object_path);"
                "self.parent.RemoveObject(self._object_path);"
                'self.EmitSignal("org.ofono.VoiceCallManager", "CallRemoved", "o", [self._object_path])',
            ),
            ("Answer", "", "", NOT_IMPLEMENTED),
        ],
    )
    obj = dbusmock.mockobject.objects[path]
    obj.parent = self
    self.calls.append(path)
    self.EmitSignal("org.ofono.VoiceCallManager", "CallAdded", "oa{sv}", [path, obj.GetProperties()])
    return path


@dbus.service.method("org.ofono.VoiceCallManager", in_signature="", out_signature="")
def HangupAll(self):
    print("XXX HangupAll", self.calls)
    for c in list(self.calls):  # needs a copy
        dbusmock.mockobject.objects[c].Hangup()
    assert self.calls == []


#  interface org.ofono.NetworkRegistration {
#    methods:
#      GetProperties(out a{sv} properties);
#      SetProperty(in s property,
#                  in v value);
#      Register();
#      GetOperators(out a(oa{sv}) operators_with_properties);
#      Scan(out a(oa{sv}) operators_with_properties);
#    signals:
#      PropertyChanged(s name,
#                      v value);
#  };
#
#  for /<modem>/operator/<CountryCode><NetworkCode>:
#  interface org.ofono.NetworkOperator {
#          methods:
#            GetProperties(out a{sv} properties);
#            Register();
#          signals:
#            PropertyChanged(s name,
#                                                  v value);
#          properties:
#        };


def get_all_operators(mock):
    return f'ret = [(m, objects[m].GetAll("org.ofono.NetworkOperator")) for m in objects if "{mock.name}/operator/" in m]'


def add_netreg_api(mock):
    """Add org.ofono.NetworkRegistration API to a mock"""

    # also add an emergency number which is not a real one, in case one runs a
    # test case against a production ofono :-)
    mock.AddProperties(
        "org.ofono.NetworkRegistration",
        {
            "Mode": "auto",
            "Status": "registered",
            "LocationAreaCode": _parameters.get("LocationAreaCode", 987),
            "CellId": _parameters.get("CellId", 10203),
            "MobileCountryCode": _parameters.get("MobileCountryCode", "777"),
            "MobileNetworkCode": _parameters.get("MobileNetworkCode", "11"),
            "Technology": _parameters.get("Technology", "gsm"),
            "Name": _parameters.get("Name", "fake.tel"),
            "Strength": _parameters.get("Strength", dbus.Byte(80)),
            "BaseStation": _parameters.get("BaseStation", ""),
        },
    )

    mock.AddObject(
        f"/{mock.name}/operator/op1",
        "org.ofono.NetworkOperator",
        {
            "Name": _parameters.get("Name", "fake.tel"),
            "Status": "current",
            "MobileCountryCode": _parameters.get("MobileCountryCode", "777"),
            "MobileNetworkCode": _parameters.get("MobileNetworkCode", "11"),
            "Technologies": [_parameters.get("Technology", "gsm")],
        },
        [
            ("GetProperties", "", "a{sv}", 'ret = self.GetAll("org.ofono.NetworkOperator")'),
            ("Register", "", "", ""),
        ],
    )

    mock.AddMethods(
        "org.ofono.NetworkRegistration",
        [
            ("GetProperties", "", "a{sv}", 'ret = self.GetAll("org.ofono.NetworkRegistration")'),
            (
                "SetProperty",
                "sv",
                "",
                'self.Set("org.ofono.NetworkRegistration", args[0], args[1]); '
                'self.EmitSignal("org.ofono.NetworkRegistration", "PropertyChanged", "sv", [args[0], args[1]])',
            ),
            ("Register", "", "", ""),
            ("GetOperators", "", "a(oa{sv})", get_all_operators(mock)),
            ("Scan", "", "a(oa{sv})", get_all_operators(mock)),
        ],
    )


#  interface org.ofono.SimManager {
#    methods:
#      GetProperties(out a{sv} properties);
#      SetProperty(in  s property,
#                  in  v value);
#      ChangePin(in  s type,
#                in  s oldpin,
#                in  s newpin);
#      EnterPin(in  s type,
#               in  s pin);
#      ResetPin(in  s type,
#               in  s puk,
#               in  s newpin);
#      LockPin(in  s type,
#              in  s pin);
#      UnlockPin(in  s type,
#                in  s pin);
#      GetIcon(in  y id,
#              out ay icon);
#    signals:
#      PropertyChanged(s name,
#                      v value);
#  };


def add_simmanager_api(self, mock):
    """Add org.ofono.SimManager API to a mock"""

    iface = "org.ofono.SimManager"
    mock.AddProperties(
        iface,
        {
            "BarredDialing": _parameters.get("BarredDialing", False),
            "CardIdentifier": _parameters.get("CardIdentifier", new_iccid(self)),
            "FixedDialing": _parameters.get("FixedDialing", False),
            "LockedPins": _parameters.get("LockedPins", dbus.Array([], signature="s")),
            "MobileCountryCode": _parameters.get("MobileCountryCode", "310"),
            "MobileNetworkCode": _parameters.get("MobileNetworkCode", "150"),
            "PreferredLanguages": _parameters.get("PreferredLanguages", ["en"]),
            "Present": _parameters.get("Present", dbus.Boolean(True)),
            "Retries": _parameters.get("Retries", dbus.Dictionary([["pin", dbus.Byte(3)], ["puk", dbus.Byte(10)]])),
            "PinRequired": _parameters.get("PinRequired", "none"),
            "SubscriberNumbers": _parameters.get("SubscriberNumbers", ["123456789", "234567890"]),
            "SubscriberIdentity": _parameters.get("SubscriberIdentity", new_imsi(self)),
        },
    )
    mock.AddMethods(
        iface,
        [
            ("GetProperties", "", "a{sv}", f'ret = self.GetAll("{iface}")'),
            (
                "SetProperty",
                "sv",
                "",
                f'self.Set("{iface}", args[0], args[1]); '
                f'self.EmitSignal("{iface}", "PropertyChanged", "sv", [args[0], args[1]])',
            ),
            ("ChangePin", "sss", "", ""),
            (
                "EnterPin",
                "ss",
                "",
                'correctPin = "1234"\n'
                f'iface = "{iface}"\n'
                'newRetries = self.Get(iface, "Retries")\n'
                'if args[0] == "pin" and args[1] != correctPin:\n'
                '    newRetries["pin"] = dbus.Byte(newRetries["pin"] - 1)\n'
                'elif args[0] == "pin":\n'
                '    newRetries["pin"] = dbus.Byte(3)\n'
                'self.Set(iface, "Retries", newRetries)\n'
                'self.EmitSignal(iface, "PropertyChanged", "sv", ["Retries", newRetries])\n'
                'if args[0] == "pin" and args[1] != correctPin:\n'
                "    class Failed(dbus.exceptions.DBusException):\n"
                '        _dbus_error_name = "org.ofono.Error.Failed"\n'
                '    raise Failed("Operation failed")',
            ),
            (
                "ResetPin",
                "sss",
                "",
                'correctPuk = "12345678"\n'
                f'iface = "{iface}"\n'
                'newRetries = self.Get(iface, "Retries")\n'
                'if args[0] == "puk" and args[1] != correctPuk:\n'
                '    newRetries["puk"] = dbus.Byte(newRetries["puk"] - 1)\n'
                'elif args[0] == "puk":\n'
                '    newRetries["pin"] = dbus.Byte(3)\n'
                '    newRetries["puk"] = dbus.Byte(10)\n'
                'self.Set(iface, "Retries", newRetries)\n'
                'self.EmitSignal(iface, "PropertyChanged", "sv", ["Retries", newRetries])\n'
                'if args[0] == "puk" and args[1] != correctPuk:\n'
                "    class Failed(dbus.exceptions.DBusException):\n"
                '        _dbus_error_name = "org.ofono.Error.Failed"\n'
                '    raise Failed("Operation failed")',
            ),
            ("LockPin", "ss", "", ""),
            ("UnlockPin", "ss", "", ""),
        ],
    )


#  interface org.ofono.ConnectionManager {
#    methods:
#      GetProperties(out a{sv} properties);
#      SetProperty(in  s property,
#                  in  v value);
#      AddContext(in  s type,
#                 out o path);
#      RemoveContext(in  o path);
#      DeactivateAll();
#      GetContexts(out a(oa{sv}) contexts_with_properties);
#    signals:
#      PropertyChanged(s name,
#                      v value);
#      ContextAdded(o path,
#                   v properties);
#      ContextRemoved(o path);
#  };
def add_connectionmanager_api(mock):
    """Add org.ofono.ConnectionManager API to a mock"""

    iface = "org.ofono.ConnectionManager"
    mock.AddProperties(
        iface,
        {
            "Attached": _parameters.get("Attached", True),
            "Bearer": _parameters.get("Bearer", "gprs"),
            "RoamingAllowed": _parameters.get("RoamingAllowed", False),
            "Powered": _parameters.get("ConnectionPowered", True),
        },
    )
    mock.AddMethods(
        iface,
        [
            ("GetProperties", "", "a{sv}", f'ret = self.GetAll("{iface}")'),
            (
                "SetProperty",
                "sv",
                "",
                f'self.Set("{iface}", args[0], args[1]); '
                f'self.EmitSignal("{iface}", "PropertyChanged", "sv", [args[0], args[1]])',
            ),
            ("AddContext", "s", "o", 'ret = "/"'),
            ("RemoveContext", "o", "", ""),
            ("DeactivateAll", "", "", ""),
            ("GetContexts", "", "a(oa{sv})", "ret = dbus.Array([])"),
        ],
    )


# unimplemented Modem object interfaces:
#
#  interface org.ofono.NetworkTime {
#    methods:
#      GetNetworkTime(out a{sv} time);
#    signals:
#      NetworkTimeChanged(a{sv} time);
#    properties:
#  };
#  interface org.ofono.MessageManager {
#    methods:
#      GetProperties(out a{sv} properties);
#      SetProperty(in  s property,
#                  in  v value);
#      SendMessage(in  s to,
#                  in  s text,
#                  out o path);
#      GetMessages(out a(oa{sv}) messages);
#    signals:
#      PropertyChanged(s name,
#                      v value);
#      IncomingMessage(s message,
#                      a{sv} info);
#      ImmediateMessage(s message,
#                       a{sv} info);
#      MessageAdded(o path,
#                   a{sv} properties);
#      MessageRemoved(o path);
#  };
#  interface org.ofono.CallVolume {
#    methods:
#      GetProperties(out a{sv} properties);
#      SetProperty(in  s property,
#                  in  v value);
#    signals:
#      PropertyChanged(s property,
#                      v value);
#  };



"""systemd timedated mock template

This creates the expected methods and properties of the main
org.freedesktop.timedate object. You can specify D-Bus property values like
"Timezone" or "NTP" in "parameters".
"""

# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
# of the license.

__author__ = "Iain Lane"
__copyright__ = """
(c) 2013 Canonical Ltd.
(c) 2017 - 2022 Martin Pitt <martin@piware.de>
"""

import dbus

BUS_NAME = "org.freedesktop.timedate1"
MAIN_OBJ = "/org/freedesktop/timedate1"
MAIN_IFACE = "org.freedesktop.timedate1"
SYSTEM_BUS = True


def setProperty(prop):
    return f'self.Set("{MAIN_IFACE}", "{prop}", args[0])'


def load(mock, parameters):
    mock.AddMethods(
        MAIN_IFACE,
        [
            # There's nothing this can usefully do, but provide it for compatibility
            ("SetTime", "xbb", "", ""),
            ("SetTimezone", "sb", "", setProperty("Timezone")),
            ("SetLocalRTC", "bbb", "", setProperty("LocalRTC")),
            ("SetNTP", "bb", "", setProperty("NTP") + "; " + setProperty("NTPSynchronized")),
        ],
    )

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "Timezone": parameters.get("Timezone", "Etc/Utc"),
                "LocalRTC": parameters.get("LocalRTC", False),
                "NTP": parameters.get("NTP", True),
                "NTPSynchronized": parameters.get("NTP", True),
                "CanNTP": parameters.get("CanNTP", True),
            },
            signature="sv",
        ),
    )
